from uplogic.nodes import ULParameterNode
from bpy.types import Material
import bpy


class ULGetMaterialNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.mat_name = None
        self.node_name = None
        self.OUT = self.add_output(self._get_val)

    def _get_val(self):
        material: Material = self.get_input(self.mat_name)
        node_name = self.get_input(self.node_name)
        return (
            material
            .node_tree
            .nodes[node_name]
        )



from uplogic.nodes import ULParameterNode


class ULTypeCastValue(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.to_type = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        value = self.get_input(self.value)
        return self.typecast_value(value, self.to_type)

    def typecast_value(self, value, t):
        if t == 'int':
            return int(value)
        elif t == 'bool':
            return bool(value)
        elif t == 'str':
            return str(value)
        elif t == 'float':
            return float(value)
        return value



from uplogic.nodes import ULParameterNode
from uplogic.nodes.logictree import ULLogicTree
from bge import logic


class ULTimeData(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.network: ULLogicTree = None
        self.TIME_PER_FRAME = self.add_output(self.get_time_per_frame)
        self.FPS = self.add_output(self.get_fps)
        self.TIMELINE = self.add_output(self.get_timeline)

    def get_time_per_frame(self):
        return self.network.time_per_frame

    def get_fps(self):
        tpf = self.network.time_per_frame or 1
        fps = logic.getAverageFrameRate()
        return fps if fps < 500 else 1 / tpf

    def get_timeline(self):
        return logic.getRealTime()

    def setup(self, network):
        self.network = network



from bge.types import KX_GameObject as GameObject
from uplogic.nodes import ULParameterNode


class ULChildByIndex(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.from_parent: GameObject = None
        self.index: int = None
        self.CHILD = self.add_output(self.get_child)

    def get_child(self):
        parent: GameObject = self.get_input(self.from_parent)
        index: int = self.get_input(self.index)
        if len(parent.children) > index:
            return parent.children[index]



from uplogic.nodes import ULParameterNode


class ULFormattedString(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.format_string = None
        self.value_a = None
        self.value_b = None
        self.value_c = None
        self.value_d = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        format_string = self.get_input(self.format_string)
        value_a = self.get_input(self.value_a)
        value_b = self.get_input(self.value_b)
        value_c = self.get_input(self.value_c)
        value_d = self.get_input(self.value_d)
        result = format_string.format(value_a, value_b, value_c, value_d)
        return result



from uplogic.nodes import ULParameterNode


class ULListExtend(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.list_1: list = None
        self.list_2: list = None
        self.OUT = self.add_output(self.get_done)
        self.LIST = self.add_output(self.get_list)

    def get_done(self):
        return False

    def get_list(self):
        list_1 = self.get_input(self.list_1)
        list_2 = self.get_input(self.list_2)
        return list_1.extend(list_2)



from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.utils.visualize import draw_line, draw_path, draw_box, draw_cube, draw_mesh, draw_arrow, draw_axis


class DrawNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.color = (1, 1, 1, 1)
        self.origin = Vector((0, 0, 0))
        self.target = Vector((0, 0, 0))
        self.points = []
        self.length = 1.0
        self.width = 1.0
        self.height = 1.0
        self.object = None
        self.use_volume_origin = False
        self.OUT = self.add_output(self.get_out)
        self.mode = 0
        self.modes = [
            self.draw_line,
            self.draw_arrow,
            self.draw_path,
            self.draw_cube,
            self.draw_box,
            self.draw_mesh,
            self.draw_axis
        ]

    def get_out(self):
        return self._done

    def draw_line(self):
        draw_line(
            self.get_input(self.origin),
            self.get_input(self.target),
            self.get_input(self.color)
        )

    def draw_arrow(self):
        draw_arrow(
            self.get_input(self.origin),
            self.get_input(self.target),
            self.get_input(self.color)
        )

    def draw_path(self):
        draw_path(
            self.get_input(self.points),
            self.get_input(self.color)
        )

    def draw_cube(self):
        draw_cube(
            self.get_input(self.origin),
            self.get_input(self.width),
            self.get_input(self.color),
            self.use_volume_origin
        )

    def draw_box(self):
        draw_box(
            self.get_input(self.origin),
            self.get_input(self.width),
            self.get_input(self.length),
            self.get_input(self.height),
            self.get_input(self.color),
            self.use_volume_origin
        )

    def draw_mesh(self):
        draw_mesh(
            self.get_input(self.object),
            self.get_input(self.color)
        )

    def draw_axis(self):
        draw_axis(
            self.get_input(self.object),
            self.get_input(self.length)
        )

    def evaluate(self):
        if not self.get_condition():
            return
        self.modes[self.mode]()
        self._done = True



from uplogic.nodes import ULParameterNode
from uplogic.utils import debug
import bpy
import json
import os


class ULLoadVariableDict(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.file_name = None
        self.path = ''
        self.VAR = self.add_output(self.get_var)

    def get_var(self):
        path = self.get_input(self.path)

        path = (bpy.path.abspath(path))
        os.makedirs(path, exist_ok=True)

        return self.read_from_json(path)

    def read_from_json(self, path):
        if not path.endswith('.json'):
            path = os.path.join(path, f'{self.get_input(self.file_name)}.json')
        if not os.path.isfile(path):
            debug('No Saved Variables!')
            return
        f = open(path, 'r')
        data = json.load(f)
        f.close()
        return data



from bge import render
from uplogic.input import MOUSE
from uplogic.input.mouse import MOUSE_EVENTS, LMB, RMB
from mathutils import Vector
from uplogic.utils.math import rotate2d


class HoverBehavior():
    @property
    def hover(self):
        if not self.active:
            return False
        screen_size = [render.getWindowWidth(), render.getWindowHeight()]
        mpos = Vector((MOUSE.position.x * screen_size[0], (1 - MOUSE.position.y) * screen_size[1]))
        mpos = rotate2d(mpos, self.pivot, -self._draw_angle)
        dsize = self._draw_size
        is_hover = (
            self.pos_abs[0] <= mpos.x <= self.pos_abs[0] + dsize[0] and
            self.pos_abs[1] <= mpos.y <= self.pos_abs[1] + dsize[1] and
            not self.canvas._hover_consumed
        )
        return is_hover


class MouseListener():
    """Mixin class for widgets that need mouse input.

    You will need to call `MouseListener.evaluate(self)` to update the behavior state.\n
    This data is independent of hovering, it just gives generic mouse button information.
    """

    def evaluate(self):
        self.clicked = False
        self.released = False
        self.hold = False
        if MOUSE_EVENTS[LMB].active and not self._down:
            self.on_click(self)
            self.on_press(self)
            self._down = True
            self.clicked = True
        elif not MOUSE_EVENTS[LMB].active and self._down:
            self.on_release(self)
            self._down = False
            self.released = True
        elif self._down:
            self.hold = True
            self.on_hold(self)

    def on_click(self, widget):
        pass

    def on_press(self, widget):
        pass

    def on_release(self, widget):
        pass

    def on_hold(self, widget):
        pass



from uplogic.nodes import ULActionNode
from uplogic.utils import make_valid_name


class ULExecuteSubNetwork(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self._condition = None
        self.target_object = None
        self.tree_name = None
        self._network = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def setup(self, network):
        self._network = network

    def evaluate(self):
        condition = self.get_condition()
        if condition != self._condition:
            target_object = self.get_input(self.target_object)
            tree_name = self.get_input(self.tree_name)
            tree_name = make_valid_name(tree_name)
            network = target_object.get(f'IGNLTree_{tree_name}', None)
            if network is None:
                network = self._network.install_subnetwork(
                    target_object,
                    tree_name,
                    True
                )
            network.stopped = not condition
            self._condition = condition
            self._done = True



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.utils import get_bitmask
from uplogic.utils.raycasting import raycast_mouse, RayCastData


class ULMouseRayCast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.distance = None
        self.property = None
        self.xray = None
        self.camera = None
        self.mask = get_bitmask(all=True)
        self._out_object = None
        self._out_normal = None
        self._out_point = None
        self._data = RayCastData((None, None, None, None, None, None))
        self.RESULT = self.add_output(self.get_result)
        self.OUTOBJECT = self.add_output(self.get_out_object)
        self.OUTNORMAL = self.add_output(self.get_out_normal)
        self.OUTPOINT = self.add_output(self.get_out_point)

    def get_result(self):
        return self._data.obj

    def get_out_object(self):
        return self._data.obj

    def get_out_normal(self):
        return self._data.normal

    def get_out_point(self):
        return self._data.point

    def evaluate(self):
        if not self.get_condition():
            self._data = RayCastData((None, None, None, None, None, None))
            return
        self._data = raycast_mouse(
            distance=self.get_input(self.distance),
            prop=self.get_input(self.property),
            xray=self.get_input(self.xray),
            mask=self.get_input(self.mask)

        )



from uplogic.nodes import ULParameterNode


class ULGetLightColor(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.lamp = None
        self.color = 0
        self.COLOR = self.add_output(self.get_color)

    def get_color(self):
        lamp = self.get_input(self.lamp)
        light = lamp.blenderObject.data
        return light.color



import bge, bpy


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class Engine(object):

    _pre_draw = None
    _post_draw = None
    _pre_draw_setup = None

    @classproperty
    def bge_scene(cls) -> bge.types.KX_Scene:
        bge.logic.getCurrentScene()
    
    @classproperty
    def bpy_scene(cls):
        return bpy.data.scenes[cls.bge_scene.name]

    @classproperty
    def pre_draw(cls):
        scene = cls.bpy_scene
        return (
            cls.bpy_scene.pre_draw
            if scene.game_settings.use_viewport_renderer else
            cls.bge_scene.pre_draw
        )

    @pre_draw.setter
    def pre_draw(cls, value):
        return

    @classproperty
    def post_draw(cls):
        return cls._bar

    @post_draw.setter
    def post_draw(cls, value):
        return

    @classproperty
    def pre_draw_setup(cls):
        return cls._bar

    @pre_draw_setup.setter
    def pre_draw_setup(cls, value):
        return


from uplogic.nodes import ULConditionNode


class ULLogicGateList(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.items = None
        self.list: list = None
        self.OUT = self.add_output(self.get_and_list)

        self.getters = [
            self.get_and_list,
            self.get_or_list
        ]

        self.gate = 0

    @property
    def gate(self):
        return None

    @gate.setter
    def gate(self, val):
        self.OUT.get_value = self.getters[val]

    def get_and_list(self):
        for item in self.items:
            it = self.get_input(item)
            if not it:
                return False
        return True

    def get_or_list(self):
        for item in self.items:
            it = self.get_input(item)
            if it is True:
                return True
        return False



from uplogic.nodes import ULConditionNode


class ULEvaluateProperty(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self.property_name = None
        self.operator = None
        self.mode = 0
        self.compare_value = None
        self.OUT = self.add_output(self.get_out)
        self.val = 0
        self.VAL = self.add_output(self.get_val)

    def get_out(self):
        compare_value = self.get_input(self.compare_value)
        return self.operator(self.val, compare_value)

    def get_val(self):
        return self.val

    def evaluate(self):
        game_object = self.get_input(self.game_object)
        property_name = self.get_input(self.property_name)
        obj = game_object if self.mode in [0, 'GAME'] else game_object.blenderObject
        self.val = obj.get(property_name)



from .widget import Widget
from .behaviors import HoverBehavior
from .label import Label
from .image import Image
from .image import Sprite
import gpu
from uplogic.input.mouse import MOUSE_EVENTS, LMB, RMB
from uplogic.utils import debug


class Button(Widget, HoverBehavior):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 100.],
        bg_color=(0, 0, 0, 0),
        relative={},
        border_width=1.0,
        border_color=(0, 0, 0, 0),
        hover_color=(0, 0, 0, .5),
        click_color=(0, 0, 0, 0),
        halign='left',
        valign='bottom',
        on_press=None,
        angle=0
    ):
        super().__init__(pos, size, bg_color, relative, halign=halign, valign=valign, angle=angle)
        self.hover_color = hover_color
        self.border_width = border_width
        self.border_color = border_color
        self.click_color = click_color
        self.override_color = None
        self._clicked = False
        self._released = False
        self._in_focus = False
        self._down = False
        self._hover = False
        if on_press is not None:
            self.on_press = on_press
        self.start()

    @property
    def border_color(self):
        return self._border_color

    @border_color.setter
    def border_color(self, val):
        self._border_color = list(val)

    @property
    def click_color(self):
        return self._click_color

    @click_color.setter
    def click_color(self, val):
        self._click_color = list(val)

    @property
    def hover_color(self):
        return self._hover_color

    @hover_color.setter
    def hover_color(self, val):
        self._hover_color = list(val)

    @property
    def current_color(self):
        return self.click_color if self._clicked or self._down else (self.hover_color if self._in_focus else self.bg_color).copy()

    def draw(self):
        super()._setup_draw()
        self._released = False
        gpu.state.line_width_set(self.border_width)
        gpu.state.point_size_set(self.border_width)
        col = self.override_color if self.override_color is not None else self.current_color
        col[3] *= self.opacity
        bcol = self.border_color.copy()
        bcol[3] *= self.opacity
        self._shader.uniform_float("color", col)
        self._clicked = False
        self._batch.draw(self._shader)
        self._shader.uniform_float("color", bcol)
        self._batch_line.draw(self._shader)
        self._batch_points.draw(self._shader)
        super().draw()

    def evaluate(self):
        is_hover = self.hover
        was_hover = self._hover
        if is_hover and not was_hover:
            self.on_enter(self)
        elif was_hover and not is_hover:
            self.on_exit(self)
        self._hover = self.hover
        if self._hover:
            self._in_focus = True
            self.on_hover(self)
            self.canvas._hover_consumed = True
        else:
            self._in_focus = False
        if self._in_focus and MOUSE_EVENTS[LMB].active and not self.canvas._click_consumed and not self._down:
            self.on_click(self)
            self.on_press(self)
            self._clicked = True
            self.canvas._click_consumed = True
            self._down = True
        elif not MOUSE_EVENTS[LMB].active and self._down:
            self.on_release(self)
            self._down = False
            self._released = True
        elif self._down:
            self.canvas._click_consumed = True
            self.on_hold(self)
        self.override_color = None

    def on_enter(self, widget):
        pass

    def on_exit(self, widget):
        pass

    def on_click(self, widget):
        pass

    def on_press(self, widget):
        pass

    def on_hold(self, widget):
        pass

    def on_release(self, widget):
        pass

    def on_hover(self, widget):
        pass


class LabelButton(Button, HoverBehavior):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 100.],
        relative={},
        bg_color=(0, 0, 0, 0),
        border_width=1.0,
        border_color=(0, 0, 0, 0),
        hover_color=(0, 0, 0, .5),
        click_color=(0, 0, 0, 0),
        text='',
        text_pos=[.5, .5],
        font='',
        font_size=12,
        font_color=(1, 1, 1, 1),
        line_height=1.5,
        halign='left',
        valign='bottom',
        halign_text='center',
        valign_text='center',
        on_press=None,
        angle=0
    ):
        text_rel = relative.get('font_size', False)
        self.label = Label(
            relative={'pos': True, 'font_size': text_rel},
            pos=text_pos,
            font=font,
            halign=halign_text,
            valign=valign_text,
            text=text,
            font_color=font_color,
            font_size=font_size,
            line_height=line_height
        )
        super().__init__(
            pos,
            size,
            bg_color,
            relative,
            border_width=border_width,
            border_color=border_color,
            hover_color=hover_color,
            click_color=click_color,
            halign=halign,
            valign=valign,
            angle=angle,
            on_press=on_press
        )
        self.add_widget(self.label)
        self._in_focus = False
        self.start()

    @property
    def text(self):
        return self.label.text

    @text.setter
    def text(self, val):
        self.label.text = val

    @property
    def text_pos(self):
        return self.label.pos

    @text_pos.setter
    def text_pos(self, val):
        self.label.pos = val

    @property
    def font(self):
        return self.label.font

    @font.setter
    def font(self, val):
        self.label.font = val

    @property
    def font_size(self):
        return self.label.font_size

    @text.setter
    def font_size(self, val):
        self.label.font_size = val

    @property
    def font_color(self):
        return self.label.font_color

    @font_color.setter
    def font_color(self, val):
        self.label.font_color = val

    @property
    def line_height(self):
        return self.label.line_height

    @line_height.setter
    def line_height(self, val):
        self.label.line_height = val

    @property
    def halign_text(self):
        return self.label.halign

    @halign_text.setter
    def halign_text(self, val):
        self.label.halign = val

    @property
    def valign_text(self):
        return self.label.valign

    @valign_text.setter
    def valign_text(self, val):
        self.label.valign = val


class ImageButton(Button, HoverBehavior):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 100.],
        bg_color=(0, 0, 0, 0),
        border_color=(0, 0, 0, 0),
        hover_color=(0, 0, 0, .0),
        click_color=(0, 0, 0, 0),
        relative={},
        halign='left',
        valign='bottom',
        on_press=None,
        texture=None,
        hover_texture=None,
        click_texture=None,
        angle=0
    ):
        super().__init__(pos, size, bg_color, relative, halign=halign, valign=valign, angle=angle, on_press=on_press, hover_color=hover_color, border_color=border_color, click_color=click_color)
        self.image = Image(relative={'size': True}, size=(1, 1), texture=texture)
        self._texture_name = texture
        self.idle_texture = texture
        self.hover_texture = hover_texture if hover_texture else texture
        self.click_texture = click_texture if click_texture else texture
        self.add_widget(self.image)
        self.start()

    @property
    def texture(self):
        return self.image.texture

    @texture.setter
    def texture(self, val):
        self._texture_name = val
        self.image.texture = val

    @property
    def current_texture(self):
        return self.click_texture if self._clicked else (
            self.hover_texture if self._hover else self.idle_texture
        )

    def evaluate(self):
        super().evaluate()
        tex = self.current_texture
        if self._texture_name != tex:
            self.texture = tex


class SpriteButton(Button, HoverBehavior):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 100.],
        bg_color=(0, 0, 0, 0),
        border_color=(0, 0, 0, 0),
        hover_color=(0, 0, 0, .0),
        relative={},
        halign='left',
        valign='bottom',
        texture=None,
        hover_texture=None,
        click_texture=None,
        idx=0,
        rows=1,
        cols=1,
        on_press=None,
        angle=0
    ):
        super().__init__(pos, size, bg_color, relative, halign=halign, valign=valign, angle=angle, on_press=on_press, hover_color=hover_color, border_color=border_color)
        self.image = Sprite(relative={'size': True}, size=(1, 1), texture=texture, rows=rows, cols=cols, idx=idx)
        self._texture_name = texture

        self.idle_texture = texture
        self.hover_texture = hover_texture if hover_texture else texture
        self.click_texture = click_texture if click_texture else texture
        self.add_widget(self.image)
        self.start()

    @property
    def texture(self):
        return self.image.texture

    @texture.setter
    def texture(self, val):
        self._texture_name = val
        self.image.texture = val

    @property
    def current_texture(self):
        return self.click_texture if self._clicked else (
            self.hover_texture if self._hover else self.idle_texture
        )

    @property
    def idx(self):
        return self.image.idx

    @idx.setter
    def idx(self, val):
        self.image.idx = val

    def evaluate(self):
        super().evaluate()
        tex = self.current_texture
        if self._texture_name != tex:
            self.texture = tex


from .widget import Widget
from uplogic.input import MOUSE
import gpu
import bge, bpy
from gpu_extras.batch import batch_for_shader
from bge import logic, render
from math import ceil
from mathutils import Vector


CURSOR = None


def remove_custom_cursor():
    scene = bge.logic.getCurrentScene()
    to_remove = []
    for f in scene.post_draw:
        if f.__name__ == '_draw_custom_cursor':
            to_remove.append(f)
    for f in to_remove:
        scene.post_draw.remove(f)


class Cursor(Widget):

    def __init__(self, texture=None, size=(30,30), offset=(0, 0), rows=1, cols=1, idx=0):
        self.offset = offset
        self.rows = rows
        self.cols = cols
        self._idx = idx
        self._to_evaluate = []
        remove_custom_cursor()
        super().__init__(MOUSE.position, size)
        self._texture = None
        self._shader = None
        self.texture = texture
        self.pos = MOUSE.position
        self._last_visible = logic.mouse.visible
        bge.logic.getCurrentScene().post_draw.append(self._draw_custom_cursor)
        self.start()

    @property
    def idx(self):
        return self._idx

    @idx.setter
    def idx(self, val):
        self._idx = val
        self._build_shader()

    @property
    def rows(self):
        return self._rows

    @rows.setter
    def rows(self, val):
        if val < 1:
            return
        self._rows = val
        self._row_height = 1 / val

    @property
    def cols(self):
        return self._cols

    @cols.setter
    def cols(self, val):
        if val < 1:
            return
        self._cols = val
        self._col_width = 1 / val

    @property
    def texture(self):
        return self._texture

    @texture.setter
    def texture(self, val):
        texture = bpy.data.images.get(val)
        if not texture:
            texture = bpy.data.images.load(val)
        texture.use_fake_user = True
        self._texture = gpu.texture.from_image(texture)

    def _build_shader(self):
        self._shader = gpu.shader.from_builtin('IMAGE_COLOR')
        screen_res = [bge.render.getWindowWidth(), bge.render.getWindowHeight()]
        mpos = [MOUSE.position.x * screen_res[0] + self.offset[0], (1 - MOUSE.position.y) * screen_res[1] + self.offset[1]]
        idx = self.idx
        col = idx % self.cols
        col_end = col + 1
        row = ceil((idx + 1) / self.cols) - 1
        row_end = row + 1
        texcoord = (
            (col_end * self._col_width, 1 - row_end * self._row_height),
            (col * self._col_width, 1 - row_end * self._row_height),
            (col_end * self._col_width, 1 - row * self._row_height),
            (col * self._col_width, 1 - row * self._row_height)
        )
        self._batch = batch_for_shader(
            self._shader, 'TRI_STRIP',
            {
                "pos": (
                    (mpos[0] + self.size[0], mpos[1] - self.size[1]),
                    (mpos[0], mpos[1] - self.size[1]),
                    (mpos[0] + self.size[0], mpos[1]),
                    mpos
                ),
                "texCoord": texcoord,
            },
        )

    @property
    def _draw_pos(self):
        return Vector((
            self.pos[0] * render.getWindowWidth(),
            (1 - self.pos[1]) * render.getWindowHeight()
        ))

    def _draw_custom_cursor(self):
        self._build_shader()
        scene = bge.logic.getCurrentScene()
        if scene.post_draw[-1] is not self._draw_custom_cursor:
            scene.post_draw.remove(self._draw_custom_cursor)
            scene.post_draw.append(self._draw_custom_cursor)
        gpu.state.blend_set('ALPHA')
        if self.show:
            self.pos = MOUSE.position
            self._shader.bind()
            self._shader.uniform_sampler("image", self.texture)
            self._batch.draw(self._shader)
        for widget in self.children:
            if widget.show:
                widget.draw()



try:
    from .button import Button, LabelButton, SpriteButton, ImageButton
    from .cursor import Cursor, remove_custom_cursor
    from .image import Image, Sprite
    from .label import Label
    from .layout import FloatLayout
    from .layout import Layout
    from .layout import RelativeLayout
    from .layout import BoxLayout
    from .layout import GridLayout
    from .layout import PolarLayout
    from .widget import Widget
    from .canvas import Canvas
    from .slider import Slider
    from .slider import FrameSlider
    from .slider import ProgressSlider
    from .path import Path
    from .worldpath import WorldPath
    from .worldpath import ObjectPath
    from .behaviors import MouseListener
    from .textinput import TextInput
except Exception:
    print('Not in game mode!')
# from .render import RenderedTexture


from uplogic.nodes import ULActionNode


class ULTranslate(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.moving_object = None
        self.local = None
        self.vect = None
        self.speed = None
        self._t = None
        self._old_values = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        print('"Translate" node is deprecated!')



from uplogic.nodes import ULActionNode


class ULStoreValue(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.value = None
        self.initialize = True
        self.condition = None
        self._stored_value = None

        # WEIRD NAMING: self.DONE = Condition, self.OUT = Stored value
        self.DONE = self.add_output(self.get_stored)
        self.OUT = self.add_output(self.get_done)

    def get_stored(self):
        return self._done

    def get_done(self):
        return self._stored_value

    def evaluate(self):
        condition = self.get_input(self.condition)
        if self.initialize:
            self.initialize = False
            condition = True
        if not condition:
            return
        self._stored_value = self.get_input(self.value)
        self._done = True



from uplogic.nodes import ULParameterNode
from bpy.types import NodeTree


class ULGetNodeAttribute(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.mat_name = None
        self.node_name = None
        self.internal = None
        self.attribute = None
        self.OUT = self.add_output(self._get_val)

    def _get_val(self):
        tree: NodeTree = self.get_input(self.mat_name)
        node_name = self.get_input(self.node_name)
        internal = self.get_input(self.internal)
        attribute = self.get_input(self.attribute)
        target = (
            tree
            .nodes[node_name]
        )
        if internal:
            target = getattr(target, internal, target)
        return getattr(target, attribute, None)



from uplogic.nodes import ULParameterNode
from uplogic.utils import FPS_FACTOR
from bge import logic


class ULFPSFactor(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.TIMEFACTOR = self.add_output(self.get_out)

    def get_out(self):
        return FPS_FACTOR()



from ..utils.visualize import draw_line
from .navigation import NavContainer
from bge.types import KX_GameObject
from bge.types import KX_NavMeshObject
from uplogic.utils import raycast
from uplogic.utils.objects import zrot_to
from uplogic.utils.constants import FPS_FACTOR
from mathutils import Vector


class Agent(NavContainer):
    """Simple AI Agent implementation that uses a Navigation Mesh.

    :param `game_object`: The game object acting as the agent.
    :param `speed`: The speed this agent will move at towards the target.
    :param `threshold`: Reach threshold for Navigation Path points. If none is set, `speed` will be used as theshold.
    :param `bevel`: Bevel distance at corners. This will cut corners outside of the navmesh.
    :param `dynamic`: Whether to move the agent using forces or pure vectors.
    :param `obstacle_mask`: Objects in this collision group will be recognized as obstacles. Set to `65535` for all objects.
    :param `height`: Z-Offset for the path calculation.
    """
    def __init__(
            self,
            game_object: KX_GameObject,
            speed: float= .1,
            threshold: float = -1,
            bevel=0.0,
            dynamic=False,
            obstacle_mask=0,
            height=0.0
        ):
        super().__init__(game_object)
        self.speed = speed
        self.navmesh = None
        self.threshold = threshold
        self.height = height
        self.obstacle_mask = obstacle_mask
        self.bevel = bevel
        self.dynamic = dynamic

    def set_navmesh(self, navmesh: KX_GameObject):
        self.navmesh = navmesh

    @property
    def position(self):
        pos = self.game_object.worldPosition.copy()
        pos.z = 0
        return pos

    @property
    def next_point(self) -> Vector:
        if self._path.points:
            if self.obstacle_mask:
                pathpoints = self._path.points
                dat = raycast(self.game_object, self.game_object.worldPosition.xy.to_3d(), pathpoints[0], distance=5, mask=self.obstacle_mask)
                if dat.obj and dat.obj.blenderObject.game.use_obstacle_create:
                    rad = dat.obj.blenderObject.game.obstacle_radius * 1.5
                    while (pathpoints[0] - dat.obj.worldPosition).length < rad:
                        self.pop()
                    
                    n = dat.normal.copy()
                    n.z = 0
                    next_direction = (pathpoints[0] - dat.obj.worldPosition).normalized()
                    normal = dat.direction.cross(Vector((0, 0, 1))).normalized()
                    normal1 = next_direction.cross(Vector((0, 0, 1))).normalized()
                    angle = dat.direction.to_2d().angle_signed(n.to_2d())
                    dir = 1 if angle > 0 else -1

                    handle_1 = dat.obj.worldPosition + normal * rad * dir
                    handle_2 = dat.obj.worldPosition + normal1 * rad * dir
                    dist_to_next_1 = (handle_1 - pathpoints[0]).length
                    dist_to_next_2 = (handle_2 - pathpoints[0]).length

                    # normal = dat.direction.cross(Vector((0, 0, 1))).normalized()
                    while (pathpoints[0] - self.game_object.worldPosition).length < (self.game_object.worldPosition - handle_1).length:
                        self.pop()
                    pathpoints.insert(0, handle_1)
                    if (pathpoints[0] - handle_2).length > self.speed and dist_to_next_1 > dist_to_next_2:
                        pathpoints.insert(1, handle_2)

            return self._path.points[0]

    def find_path(self, target: Vector, navmesh: KX_NavMeshObject = None):
        return super().find_path(self.game_object.worldPosition, target, navmesh if navmesh else self.navmesh)

    def visualize(self, color=Vector((0, 1, 0))):
        if self._path.points:
            compare = self.game_object.worldPosition.copy()
            compare.z = self.next_point.z
            draw_line(compare, self.next_point, color)
            return super().visualize(color)

    def pop(self, idx=0):
        points = self._path.points
        if not points:
            return None
        return points.pop(idx)

    def clean(self):
       while self.next_point and self.distance < .3:
           self.pop()

    @property
    def idle(self):
        return not self._path.points

    @property
    def distance(self):
        compare = self.game_object.worldPosition.copy()
        np = self.next_point
        if np is None:
            return 0
        compare.z = self.next_point.z
        return super().distance(compare)

    @property
    def direction(self):
        compare = self.game_object.worldPosition.copy()
        np = self.next_point
        if np is None:
            return Vector((0, 0, 0))
        compare.z = self.next_point.z
        return super().direction(compare)

    def lookat(self, factor=.1):
        next_point = self.next_point
        if next_point is not None:
            zrot_to(self.game_object, next_point, 1, factor)

    @property
    def threshold(self):
        if self._threshold >= 0:
            return self._threshold
        elif self.dynamic:
            return self.game_object.worldLinearVelocity.length * FPS_FACTOR()
        else:
            return self.speed

    @threshold.setter
    def threshold(self, val):
        self._threshold = val

    def move(self):
        while self._path.points and self.distance < self.threshold:
            self.pop()
        if not self._path.points:
            return
        if self.dynamic:
            self.game_object.applyForce(self.direction * self.speed)
        else:
            self.game_object.applyMovement(self.direction * self.speed)


from .widget import Widget
import gpu
from bge import render
from mathutils import Vector
from ..utils.math import rotate2d
from ..events import schedule, ScheduledEvent


class Layout(Widget):
    '''The Layout class allows you to arrange widgets in a specified area.

    :param `pos`: Initial position of this widget in either pixels or factor.
    :param `size`: Initial size of this widget in either pixels or factor.
    :param `bg_color`: Color to draw in the area of the widget.
    :param `relative`: Whether to use pixels or factor for size or pos; example: `{'pos': True, 'size': True}`.
    :param `border_width`: Width (in pixels) of the border drawn around the area of the widget.
    :param `border_color`: Color to use for drawing the border.
    :param `halign`: Horizontal alignment of the widget, can be (`left`, `center`, `right`).
    :param `valign`: Vertical alignment of the widget, can be (`bottom`, `center`, `top`).
    :param `angle`: Rotation in degrees of this widget around the pivot defined by the alignment.
    '''

    def __init__(
        self,
        pos: list = [0., 0.],
        size: list = [100., 100.],
        bg_color: list = [0, 0, 0, 0],
        relative: dict = {},
        border_width: int = 1,
        border_color: list = [0, 0, 0, 0],
        halign: str = 'left',
        valign: str = 'bottom',
        angle=0
    ):
        self.border_width = border_width
        self.border_color = border_color
        super().__init__(pos, size, bg_color, relative, halign=halign, valign=valign, angle=angle)
        self.start()

    @property
    def border_color(self):
        return self._border_color

    @border_color.setter
    def border_color(self, color):
        self._border_color = list(color)

    @property
    def border_width(self):
        return self._border_width

    @border_width.setter
    def border_width(self, val):
        if val < 1:
            val = 1
        self._border_width = int(val)

    def draw(self):
        super()._setup_draw()
        gpu.state.line_width_set(self.border_width)
        gpu.state.point_size_set(self.border_width)
        col = self.bg_color.copy()
        col[3] *= self.opacity
        bcol = self.border_color.copy()
        bcol[3] *= self.opacity
        self._shader.uniform_float("color", col)
        self._batch.draw(self._shader)
        self._shader.uniform_float("color", bcol)
        self._batch_line.draw(self._shader)
        self._batch_points.draw(self._shader)
        super().draw()


class RelativeLayout(Layout):
    '''The RelativeLayout allows you to place widgets relative to the Layouts coordinates.

    :param `pos`: Initial position of this widget in either pixels or factor.
    :param `size`: Initial size of this widget in either pixels or factor.
    :param `bg_color`: Color to draw in the area of the widget.
    :param `relative`: Whether to use pixels or factor for size or pos; example: `{'pos': True, 'size': True}`.
    :param `border_width`: Width (in pixels) of the border drawn around the area of the widget.
    :param `border_color`: Color to use for drawing the border.
    :param `halign`: Horizontal alignment of the widget, can be (`left`, `center`, `right`).
    :param `valign`: Vertical alignment of the widget, can be (`bottom`, `center`, `top`).
    :param `angle`: Rotation in degrees of this widget around the pivot defined by the alignment.
    '''
    @property
    def clipping(self):
        return [
            0,
            render.getWindowWidth(),
            render.getWindowHeight(),
            0
        ]


class FloatLayout(Layout):
    '''The FloatLayout allows you to place widgets in Canvas space.

    :param `pos`: Initial position of this widget in either pixels or factor.
    :param `size`: Initial size of this widget in either pixels or factor.
    :param `bg_color`: Color to draw in the area of the widget.
    :param `relative`: Whether to use pixels or factor for size or pos; example: `{'pos': True, 'size': True}`.
    :param `border_width`: Width (in pixels) of the border drawn around the area of the widget.
    :param `border_color`: Color to use for drawing the border.
    :param `halign`: Horizontal alignment of the widget, can be (`left`, `center`, `right`).
    :param `valign`: Vertical alignment of the widget, can be (`bottom`, `center`, `top`).
    :param `angle`: Rotation in degrees of this widget around the pivot defined by the alignment.
    '''

    @property
    def pos_abs(self):
        return [0, 0]


class ArrangedLayout(Layout):
    """Metaclass"""
    
    @property
    def arrange_event(self) -> ScheduledEvent:
        return getattr(self, '_arrange_evt', None)

    @arrange_event.setter
    def arrange_event(self, val):
        if self.arrange_event:
            self.arrange_event.cancel()
        self._arrange_evt = val

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, val):
        if self.parent is not val and self.parent:
            self.parent.remove_widget(self)
        if self.use_clipping is None:
            self.use_clipping = val.use_clipping
        self._parent = val
        self.pos = self.pos
        self.size = self.size
        for c in self.children:
            c.parent = c.parent
        self.arrange()

    @property
    def show(self):
        return self._show

    @show.setter
    def show(self, val):
        if val != self._show:
            self._show = val
            if val:
                self._rebuild = True
                for child in self.children:
                    child.pos = child.pos
        self.arrange()

    def add_widget(self, widget):
        super().add_widget(widget)
        self.arrange()
    
    def remove_widget(self, widget):
        super().remove_widget(widget)
        self.arrange()

    def arrange(self):
        raise NotImplementedError


class BoxLayout(ArrangedLayout):
    '''The BoxLayout allows you to automatically arrange widgets in a row or column.

    :param `orientation`: Whether to arrange widgets horizontally or vertically; Can be (`'horizontal'`, `'vertical'`).
    :param `pos`: Initial position of this widget in either pixels or factor.
    :param `size`: Initial size of this widget in either pixels or factor.
    :param `bg_color`: Color to draw in the area of the widget.
    :param `relative`: Whether to use pixels or factor for size or pos; example: `{'pos': True, 'size': True}`.
    :param `border_width`: Width (in pixels) of the border drawn around the area of the widget.
    :param `border_color`: Color to use for drawing the border.
    :param `spacing`: Pixels in between child widgets.
    :param `halign`: Horizontal alignment of the widget, can be (`left`, `center`, `right`).
    :param `valign`: Vertical alignment of the widget, can be (`bottom`, `center`, `top`).
    :param `angle`: Rotation in degrees of this widget around the pivot defined by the alignment.
    '''
    def __init__(
            self,
            orientation: str = 'horizontal',
            pos: list = [0, 0],
            size: list = [100, 100],
            bg_color: list = (0, 0, 0, 0),
            relative: dict = {},
            border_width: int = 1,
            border_color: list = (0, 0, 0, 0),
            spacing: int = 0,
            halign: str = 'left',
            valign: str = 'bottom',
            angle=0
        ):
        self.orientation = orientation
        self.spacing = spacing
        self.children_align = ['left', 'bottom']
        super().__init__(pos, size, bg_color, relative, border_width, border_color, halign=halign, valign=valign, angle=angle)
        self.use_clipping = False
        self.start()

    def arrange(self):
        '''Arrange the widgets according to the specified orientation.'''
        dsize = self._draw_size
        arrange_factor = {
            'left': 0,
            'center': .5,
            'right': 1,
            'top': 0,
            'bottom': 1
        }
        xalign = self.children_align[0]
        yalign = self.children_align[1]
        if self.orientation == 'horizontal':
            offset = 0
            for widget in self.children:
                widget.halign = xalign
                widget.valign = yalign
                widget.relative['pos'] = False
                widget.pos = [offset, dsize[1] - (widget._draw_size[1] * arrange_factor[yalign])]
                offset += widget._draw_size[0] + self.spacing
        if self.orientation == 'vertical':
            offset = dsize[1]
            for widget in self.children:
                widget.halign = xalign
                widget.valign = yalign
                offset -= widget._draw_size[1]
                widget.relative['pos'] = False
                widget.pos = [arrange_factor[xalign] * dsize[0], offset]
                offset -= self.spacing


class GridLayout(BoxLayout):
    '''The GridLayout allows you automatically arrange widgets in a grid.

    :param `orientation`: Whether to arrange widgets horizontally or vertically; Can be (`'horizontal'`, `'vertical'`).
    :param `pos`: Initial position of this widget in either pixels or factor.
    :param `size`: Initial size of this widget in either pixels or factor.
    :param `bg_color`: Color to draw in the area of the widget.
    :param `relative`: Whether to use pixels or factor for size or pos; example: `{'pos': True, 'size': True}`.
    :param `border_width`: Width (in pixels) of the border drawn around the area of the widget.
    :param `border_color`: Color to use for drawing the border.
    :param `spacing`: Pixels in between child widgets.
    :param `cols`: How many columns this layout should have.
    :param `rows`: How many rows this column should have.
    :param `halign`: Horizontal alignment of the widget, can be (`left`, `center`, `right`).
    :param `valign`: Vertical alignment of the widget, can be (`bottom`, `center`, `top`).
    :param `angle`: Rotation in degrees of this widget around the pivot defined by the alignment.
    '''

    def __init__(
        self,
        orientation: str = 'vertical',
        pos: list = [0, 0],
        size: list = [100, 100],
        bg_color: list = [0, 0, 0, 0],
        relative: dict = {},
        border_width: int = 1,
        border_color: list = [0, 0, 0, 0],
        spacing: int = 0,
        cols: int = 2,
        rows: int = 2,
        halign: str = 'left',
        valign: str = 'bottom',
        angle=0
    ):
        super().__init__(
            orientation,
            pos,
            size,
            bg_color,
            relative,
            border_width,
            border_color,
            spacing,
            halign,
            valign,
            angle=angle
        )
        self.rows = rows
        self.cols = cols
        self.start()

    def add_widget(self, widget):
        max = self.rows * self.cols
        if len(self.children) < max:
            super().add_widget(widget)
            self.arrange()

    def arrange(self):
        dsize = self._draw_size
        idx = 0
        _widget_sizes = []
        _offset_y = 0
        if self.orientation == 'horizontal':
            row = 0
            offset = 0
            for widget in self.children:
                offset_y = _offset_y + (self.spacing if row else 0)
                widget.relative['pos'] = False
                wsize = widget._draw_size
                widget.pos = [offset, dsize[1] - wsize[1] - offset_y]
                _widget_sizes.append(widget._draw_pos[1])
                offset += wsize[0] + self.spacing
                idx += 1
                if idx >= self.cols:
                    _offset_y = min(_widget_sizes)
                    _widget_sizes = []
                    idx = 0
                    row += 1
                    offset = 0
        if self.orientation == 'vertical':
            col = 0
            offset = dsize[1]
            for widget in self.children:
                offset_x = (self._draw_size[0] / (self.cols) + self.spacing) * col
                offset -= widget._draw_size[0]
                widget.relative['pos'] = False
                widget.pos = [offset_x, offset]
                offset -= self.spacing
                idx += 1
                if idx >= self.rows:
                    idx = 0
                    col += 1
                    offset = dsize[1]


class PolarLayout(ArrangedLayout):
    '''The Polar Layout allows you automatically arrange widgets in a circular fashion.

    :param `pos`: Initial position of this widget in either pixels or factor.
    :param `relative`: Whether to use pixels or factor for size or pos; example: `{'pos': True}`.
    :param `starting_angle`: Position angle of the first widget. 0 is to the right, 90 is up, 180 is left, 270 is down.
    :param `angle`: Rotation in degrees of this widget around the pivot defined by the alignment.
    '''

    def __init__(
            self,
            pos: list = [0, 0],
            relative: dict = {},
            starting_angle: str = 0,
            radius: int = 100,
            angle: float = 0
        ):
        self._starting_angle = starting_angle
        self._radius = radius
        super().__init__(
            pos,
            (0, 0),
            (0, 0, 0, 0),
            relative,
            0,
            (0, 0, 0, 0),
            'center',
            'center',
            angle
        )
        self.starting_angle = starting_angle
        self.radius = radius
        self.start()

    @property
    def starting_angle(self):
        return self._starting_angle

    @starting_angle.setter
    def starting_angle(self, val):
        self._starting_angle = val
        self.arrange()

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, val):
        self._radius = val
        self.arrange()

    def add_widget(self, widget):
        super().add_widget(widget)
        w = self.children[-1]
        w.pos = Vector((self.radius, 0))

    def arrange(self):
        if len(self.children) == 0:
            return
        step = 360 / len(self.children)
        _angle = self.starting_angle
        pos = Vector((self.radius, 0))
        for widget in self.children:
            widget.relative['pos'] = False
            widget.pos = rotate2d(pos, (0, 0), _angle)
            _angle += step



from uplogic.nodes import ULConditionNode


class ULNone(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.checked_value = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self.get_input(self.checked_value) is None



from bge.types import KX_Camera
from gpu.types import GPUOffScreen
from uplogic.ui.widget import Widget
from gpu_extras.presets import draw_texture_2d
import bge
import bpy
import gpu


class Camera(Widget):

    def __init__(self, pos=[0, 0], size=(100, 100), relative={}, camera=None, halign='left', valign='bottom') -> None:
        self.camera = camera
        Widget.__init__(self, pos, size, (0, 0, 0, 0), relative, halign, valign, 0)
        self.start()

    @property
    def parent(self):
        """The widget whose position and size to use relatively."""
        return self._parent

    @parent.setter
    def parent(self, val):
        if self.parent is not val and self.parent:
            self.parent.remove_widget(self)
        if self.use_clipping is None:
            self.use_clipping = val.use_clipping
        self._parent = val
        self.pos = self.pos
        self.size = self.size
        self.camera.useViewport = val is not None and val.show and self.show
        self._build_shader()

    @property
    def camera(self) -> KX_Camera:
        return self._camera

    @camera.setter
    def camera(self, val):
        self._camera = val

    def draw(self):
        super()._setup_draw()
        self.camera.useViewport = self.parent is not None and self.parent.show and self.show
        super().draw()

    def _build_shader(self):
        if self.camera is None:
            return
        pos = self._draw_pos
        size = self._draw_size
        self.camera.setViewport(
            round(pos[0]),
            round(pos[1]),
            round(pos[0] + size[0]),
            round(pos[1] + size[1])
        )



from uplogic.nodes import ULActionNode
from uplogic.utils import make_unique_light


class ULMakeUniqueLight(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.light = None
        self._light = None
        self.OUT = self.add_output(self.get_done)
        self.LIGHT = self.add_output(self.get_light)

    def get_done(self):
        return self._done

    def get_light(self):
        return self._light

    def evaluate(self):
        if not self.get_condition():
            return
        self._light = make_unique_light(self.get_input(self.light))
        self._done = True



from uplogic.nodes import ULParameterNode


class DictGetKeysNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.dict = None
        self.KEYS = self.add_output(self.get_keys)

    def get_keys(self):
        return list(self.get_input(self.dict).keys())



from uplogic.nodes import ULOutSocket, ULParameterNode


class ULGetProperty(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.property_name = None
        self.mode = 0
        self.OUT = self.add_output(self.get_property)

    def get_property(self):
        game_object = self.get_input(self.game_object)
        property_name = self.get_input(self.property_name)
        props = (
            [p[0] for p in game_object.blenderObject.items()]
            if self.mode else
            game_object.getPropertyNames()
        )
        if property_name in props:
            obj = game_object if self.mode in [0, 'GAME'] else game_object.blenderObject
            return obj.get(property_name)
        game_object[property_name] = None



from uplogic.nodes import ULActionNode
from bpy.types import NodeTree


class ULSetNodeSocket(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.tree_name = None
        self.node_name = None
        self.input_slot = None
        self.value = None
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        tree: NodeTree = self.get_input(self.tree_name)
        node_name = self.get_input(self.node_name)
        input_slot = self.get_input(self.input_slot)
        value = self.get_input(self.value)
        (
            tree
            .nodes[node_name]
            .inputs[input_slot]
            .default_value
        ) = value
        self._done = True



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULMouseData(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.invert_y = False
        self.MX = self.add_output(self.getmx)
        self.MY = self.add_output(self.getmy)
        self.MDX = self.add_output(self.getmdx)
        self.MDY = self.add_output(self.getmdy)
        self.MDWHEEL = self.add_output(self.getmdwheel)
        self.MXY0 = self.add_output(self.getmxyz)
        self.MDXY0 = self.add_output(self.getmdxyz)

    def getmx(self):
        return self.network.mouse.position[0]

    def getmy(self):
        y = self.network.mouse.position[1]
        return (
            1-y if self.invert_y else y
        )

    def getmdx(self):
        return self.network.mouse.movement[0]

    def getmdy(self):
        y = self.network.mouse.movement[1]
        return (
            -y if self.invert_y else y
        )

    def getmdwheel(self):
        return self.network.mouse.wheel

    def getmxyz(self):
        pos = self.network.mouse.position
        y = pos[1]
        return Vector((
            pos[0],
            1 - y if self.invert_y else y,
        ))

    def getmdxyz(self):
        movement = self.network.mouse.movement
        y = movement[1]
        return Vector((
            movement[0],
            -y if self.invert_y else y,
        ))



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.utils.objects import add_object


class ULAddObject(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.name = None
        self.reference = None
        self.life = None
        self.full_copy = False
        self.obj = False
        self.OBJ = self.add_output(self._get_obj)
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def _get_obj(self):
        return self.obj

    def evaluate(self):
        condition = self.get_condition()
        if not condition:
            return
        life = self.get_input(self.life)
        name = self.get_input(self.name)
        full_copy = self.get_input(self.full_copy)
        reference = self.get_input(self.reference)
        scene = logic.getCurrentScene()
        # self.obj = scene.addObject(name, reference, life, full_copy)
        self.obj = add_object(name, reference, life, full_copy)
        self._done = True



from uplogic.nodes import ULParameterNode
from bge.types import KX_GameObject


class ULChildByName(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.from_parent = None
        self.child = None
        self.CHILD = self.add_output(self.get_child)

    def get_child(self):
        parent: KX_GameObject = self.get_input(self.from_parent)
        child: KX_GameObject = self.get_input(self.child)
        if isinstance(child, str):
            return parent.childrenRecursive.get(child, None)
        return parent.childrenRecursive.get(child.name, None)



from .widget import Widget
import blf
from uplogic.utils.math import clamp
from uplogic.ui.label import Label
from uplogic.ui.layout import Layout
from bge import logic
from bge import events
import math
from uplogic.events import schedule


class TextInput(Label):
    '''Widget for displaying text

    :param `pos`: Initial position of this widget in either pixels or factor.
    :param `relative`: Whether to use pixels or factor for size or pos; example: `{'pos': True, 'size': True}`.
    :param `text`: Initial text for this label.
    :param `font`: Font name.
    :param `font_color`: Color in RGBA.
    :param `font_size`: Font size in pt or factor.
    :param `line_height`: Total line height relative to character size (1 is same as character height).
    :param `shadow`: Draw a shadow behind the text.
    :param `shadow_offset`: Relative position of the shadow in px.
    :param `shadow_color`: Shadow color in RGBA.
    :param `halign`: Horizontal alignment of the widget, can be (`left`, `center`, `right`).
    :param `valign`: Vertical alignment of the widget, can be (`bottom`, `center`, `top`).
    :param `wrap`: Split lines that are too long for the containing widget.
    :param `multiline`: Allow paragraphs. If `False`, pressing enter will exit edit mode.
    :param `angle`: Rotation in degrees of this widget around the pivot defined by the alignment.'''

    def __init__(
        self,
        pos=[0, 0],
        relative={},
        text='',
        font='',
        font_color=[1., 1., 1., 1.],
        font_size=12,
        line_height=1.5,
        shadow=False,
        shadow_offset=[1, -1],
        shadow_color=[.0, .0, .0, 1.0],
        halign='left',
        valign='bottom',
        wrap=False,
        multiline=False,
        angle=0
    ):
        self.lines = None
        self.cursor_evt = None
        self.cursor = Layout(bg_color=(1, 1, 1, 1), size=(0, 0))
        self.cursor_flash_time = .5
        Label.__init__(self, pos, relative, text, font, font_color, font_size, line_height, shadow, shadow_offset, shadow_color, halign, valign, wrap, angle)
        self.add_widget(self.cursor)
        self._key_evts = {}
        events = logic.keyboard.inputs.copy()
        for evt in events.values():
            self._key_evts[evt] = 0.0
        self.multiline = multiline
        self.edit = False
        self._index = 0
        self.line_index = 0
        self.character_index = 0
        self.start()

    def on_enter(self):
        self.edit = False

    @property
    def edit(self):
        return self._edit

    @edit.setter
    def edit(self, val):
        if val:
            if not self._edit:
                self.index = len(self.text)
            self.cursor.show = True
            if self.cursor_evt:
                self.cursor_evt.cancel()
            self.cursor_evt = schedule(self._cursor_visible, self.cursor_flash_time)
        else:
            self.cursor.show = False
            if self.cursor_evt:
                self.cursor_evt.cancel()
        self._edit = val

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, val):
        val = clamp(val, 0, len(self.text))
        character_index = val
        line_index = 0
        for line in self.lines:
            if character_index > len(line) and (line != '' or line_index != len(self.lines)):
                line_index += 1
                character_index -= len(line) + 1
            else:
                break
        self._index = val
        self.line_index = line_index
        self.character_index = character_index

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, val):
        if self.cursor_evt:
            self.cursor_evt.cancel()
        self.cursor.show = False
        self._cursor_visible()
        self._text = str(val)

    def move_cursor_to_end(self):
        line_length = len(self.lines[clamp(self.line_index, 0, len(self.lines) - 1)])
        self.index = len(self.text) #line_length - self.character_index
        self.edit = True

    def write(self, text):
        self.text += text
        self.index = len(self.text)

    @property
    def _current_char_pos(self):
        self.index = self.index
        relative = self.relative.get('font_size', False)
        fontsize = self.parent._draw_size[1] * self.font_size if relative else self.font_size
        blf.size(self.font, fontsize)
        if len(self.lines) > 0:
            text = self.lines[self.line_index]
            text = text[0 : self.character_index]
            dim = blf.dimensions(self.font, text)
            # lines = len(self.lines) or 1
            return [dim[0], blf.dimensions(self.font, 'A')[1] * (self.line_index + 1) * self.line_height]
        else:
            return (0, 0)

    def _cursor_visible(self):
        self.cursor.show = not self.cursor.show
        self.cursor_evt = schedule(self._cursor_visible, self.cursor_flash_time)

    def _set_cursor_position(self):
        dsize = self._current_char_pos
        dim = blf.dimensions(self.font, self.lines[self.line_index])
        # charsize = blf.dimensions(self.font, 'A')
        if self.text_halign == 'center':
            dsize[0] -= dim[0] * .5
        if self.text_halign == 'right':
            dsize[0] -= dim[0]
        if self.text_valign == 'center':
            dsize[1] *= -.5
        if self.text_valign == 'top':
            dsize[1] *= -1
        self.cursor.pos[0] = dsize[0]
        self.cursor.pos[1] = dsize[1] # -charsize[1] * self.line_height * (self.line_index + 1)
        self.cursor.height = self.font_size
        self.cursor.width = 1

    def _listen(self):
        left_shift = logic.keyboard.inputs[events.LEFTSHIFTKEY].active
        right_shift = logic.keyboard.inputs[events.RIGHTSHIFTKEY].active
        keyboard_events = logic.keyboard.inputs.copy()
        tpf = (1 / (logic.getAverageFrameRate() or 0.01))

        for evt in keyboard_events.values():

            text = self.text
            if evt.active and self._key_evts[evt] == 0.0:
                self._key_evts[evt] += tpf
                if evt.type == 59:
                    self.text = text[:self.index][:-1] + text[self.index:]  # Backspace:
                    self.index -= 1
                    return
                if not self.multiline and evt.type == 7:
                    self.on_enter()
                    return
                if evt.type == 69:  # Arrow Left
                    self.index -= 1
                    self.edit = True
                if evt.type == 71:  # Arrow Right
                    self.index += 1
                    self.edit = True
                if len(self.lines) <= 0:
                    return
                line_length = len(self.lines[clamp(self.line_index, 0, len(self.lines) - 1)])
                if evt.type == 110:  # Home
                    self.index -= line_length - (line_length - self.character_index)
                    self.edit = True
                if evt.type == 113:  # End
                    self.index += line_length - self.character_index
                    self.edit = True
                # if evt.type == 70:  # Arrow Down
                # if evt.type == 72:  # Arrow Up
                char = events.EventToCharacter(evt.type, left_shift or right_shift)
                if not char:
                    continue
                self.text = text[:self.index] + char + text[self.index:]
                self.index += 1
            elif evt.active:
                self._key_evts[evt] += tpf
                if self._key_evts[evt] > .4:
                    if evt.type == 59:
                        self.text = text[:self.index - 1][:-1] + text[self.index:]  # Backspace:
                        self.index -= 1
                        return
                    if not self.multiline and evt.type == 7:  # Enter
                        self.edit = False
                        return
                    if evt.type == 69:  # Arrow Left
                        self.index -= 1
                        self.edit = True
                    if evt.type == 71:  # Arrow Right
                        self.index += 1
                        self.edit = True
                    text = self.text
                    char = events.EventToCharacter(evt.type, left_shift or right_shift)
                    if not char:
                        continue
                    self.text = text[:self.index] + char + text[self.index:]
                    self.index += 1
            elif evt.inactive:
                self._key_evts[evt] = 0.0

    def move_cursor(self, line, position):
        raise NotImplementedError

    def draw(self):
        if self.edit:
            self._listen()
        super().draw()
        self._set_cursor_position()


from uplogic.animation import ActionSystem
from uplogic.data import GlobalDB
from uplogic.animation import Action
from uplogic.events import receive
from uplogic.nodes import ULActionNode
from uplogic.utils.math import clamp
from bpy.types import Action as BPYAction


class ULPlayAction(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.action_name = None
        self.stop_anim = None
        self.frames = None
        self.start_frame = None
        self.end_frame = None
        self.layer = None
        self.priority = None
        self.play_mode = None
        self.layer_weight = None
        self.old_intensity = None
        self.speed = None
        self.old_speed = None
        self.blendin = None
        self.blend_mode = None
        self.in_use = False
        self._started = False
        self._running = False
        self._finished = False
        self._action = None
        self.action_evt = None
        self.act_system = self.get_act_system()
        self.STARTED = self.add_output(self._get_started)
        self.FINISHED = self.add_output(self._get_finished)
        self.RUNNING = self.add_output(self._get_running)
        self.FRAME = self.add_output(self._get_frame)
        
    def get_act_system(self):
        act_systems = GlobalDB.retrieve('uplogic.animation')
        if act_systems.check('default'):
            return act_systems.get('default')
        else:
            return ActionSystem('default')

    def on_start(self):
        self._started = True

    def _get_started(self):
        return self._started
    
    def on_finish(self):
        self._finished = True

    def _get_finished(self):
        return self._finished

    def _get_running(self):
        if self._action:
            return self._action.is_playing
        return False

    def _get_frame(self):
        if self._action:
            return self._action.frame
        return False
    
    def reset(self):
        self._finished = False
        self._started = False
        return super().reset()

    def evaluate(self):
        condition = self.get_condition()
        intensity = clamp(self.get_input(self.layer_weight))
        speed = self.get_input(self.speed)
        layer = self.get_input(self.layer)
        game_object = self.get_input(self.game_object)
        layer_action: Action = self.act_system.get_layer(game_object, layer)
        if layer_action is not self._action:
            self._action = layer_action
        action = self._action
        has_action = action is not None
        play_mode = self.get_input(self.play_mode)
        self.action_evt = receive(self._action)
        if not condition:
            if has_action:
                action.speed = speed
                action.intensity = intensity
                if self._action.finished:
                    self._action = None
                    self.in_use = False
                elif play_mode > 2:
                    self._action.remove()
                    self._action = None
                    self.in_use = False
            return
        bpy_action: BPYAction = self.get_input(self.action_name)
        if layer_action and layer_action.name == bpy_action:
            layer_action.speed = speed
            layer_action.intensity = intensity
            return
        if self.in_use and has_action:
            return
        if has_action and action.finished:
            self._action = None
        start_frame = self.get_input(self.start_frame)
        end_frame = self.get_input(self.end_frame)
        # priority = self.get_input(self.priority)
        blendin = self.get_input(self.blendin)
        blend_mode = self.get_input(self.blend_mode)
        if play_mode > 2:
            play_mode -= 3
        if speed <= 0:
            speed = 0.01

        self._action = Action(
            game_object=game_object,
            action_name=bpy_action.name,
            start_frame=start_frame,
            end_frame=end_frame,
            layer=layer,
            # priority=priority,
            blendin=blendin,
            play_mode=play_mode,
            speed=speed,
            intensity=intensity,
            blend_mode=blend_mode,
            on_start=self.on_start
        )
        self._action.on_finish = self.on_finish
        self.in_use = True



from bge import logic

def alpha_move(a, b, fac):
    if a < b:
        return a + fac
    elif a > b:
        return a - fac
    else:
        return a


_loaded_userlogic_files = {}


def load_user_logic(module_name):
    full_path = logic.expandPath(
        "//bgelogic/cells/{}.py".format(module_name)
    )
    loaded_value = _loaded_userlogic_files.get(full_path)
    if loaded_value:
        return loaded_value
    import sys
    python_version = sys.version_info
    major = python_version[0]
    minor = python_version[1]
    if (major < 3) or (major == 3 and minor < 3):
        import imp
        loaded_value = imp.load_source(module_name, full_path)
    elif (major == 3) and (minor < 5):
        from importlib.machinery import SourceFileLoader
        loaded_value = SourceFileLoader(module_name, full_path).load_module()
    else:
        import importlib.util
        spec = importlib.util.spec_from_file_location(module_name, full_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        loaded_value = module
    _loaded_userlogic_files[module_name] = loaded_value
    return loaded_value


class LoopList(list):
    pass


class ULLogicBase(object):
    def get_value(self, to_node=None):
        pass


class ULLogicContainer(ULLogicBase):

    def __init__(self):
        self._uid = None
        self._value = None
        self._children = []
        self._done = False
        self.network = None

    def get_value(self, to_node=None):
        return self._value

    def _set_value(self, value):
        self._value = value

    def setup(self, network):
        """
        This is called by the network once, after all the
        cells have been loaded into the tree.
        :return: None
        """
        pass

    def stop(self, network):
        pass

    def reset(self):
        """
        Resets the status of the cell.
        A cell may override this to reset other states
        or to keep the value if evaluation is required
        to happen only once (or never at all)
        :return:
        """
        pass

    def evaluate(self):
        """
        A logic cell implements this method to do its job. The network
        evaluates a cell. When that happens, the cell is
         removed from the update queue.
        :return:
        """
        raise NotImplementedError(
            "{} doesn't implement evaluate".format(self.__class__.__name__)
        )

    def _skip_evaluate(self):
        return

    def deactivate(self):
        self.evaluate = self._skip_evaluate


###############################################################################
# Socket
###############################################################################


class Output(ULLogicBase):
    def __init__(self, node, value_getter):
        self.node = node
        self.result = None
        self._value_getter = value_getter

    @property
    def result(self):
        return self._result
    
    @result.setter
    def result(self, val):
        # if self.node.loop_mode and val is None:
        #     self._result = LoopList()
        #     return
        self._result = val

    def _value_getter(self):
        pass

    def get_value(self, to_node=None):
        self.node.fetched = True
        result = self.result
        if result is None:
            result = self._value_getter()
            if self.node.loop_mode:
                result = []
                for x in range(self.node.loop_size):
                    self.node.loop_idx = x
                    result.append(self._value_getter())
                result = LoopList(result)
            self.result = result
        return result


class ULOutSocket(Output):
    pass


###############################################################################
# Basic Cells
###############################################################################


def results(*names):

    def deco(cls):
        for name in names:

            def getResult(self, name=name):
                value = self.results.get(name, None)
                if self.loop_mode and value is not None:
                    return value[self.loop_idx]
                return value

            def setResult(self, value, name=name):
                if self.loop_mode:
                    set_val = value
                    value = self.results.get(name, LoopList())
                    value.append(set_val)
                self.results[name] = value

            prop = property(getResult, setResult)
            setattr(cls, name, prop)

        return cls
    return deco


class ULLogicNode(ULLogicContainer):

    def __init__(self):
        self.loop_mode = False
        self.loop_idx = 0
        self.loop_size = 0
        self.condition = False
        self.results = {}
        self.outputs = []
        self.fetched = False
        self.get_input = self._get_input
        super().__init__()

    @property
    def loop_mode(self):
        return self._loop_mode

    @loop_mode.setter
    def loop_mode(self, val):
        self._loop_mode = val
        self.get_input = self._get_input_loop if val else self._get_input

    def reset(self):
        super().reset()
        self.results = {}
        self.loop_mode = False
        self.loop_idx = 0
        self.get_input = self._get_input
        for o in self.outputs:
            o.result = None

    def add_output(self, getter):
        o = Output(self, getter)
        self.outputs.append(o)
        return o

    def _get_input(self, param):
        if isinstance(param, ULLogicBase):
            return param.get_value(self)
        return param

    def _get_input_loop(self, param):
        if isinstance(param, ULLogicBase):
            param.node.loop_mode = True
            param.node.loop_size = self.loop_size
            param.node.loop_idx = self.loop_idx
            param = param.get_value(self)
        if isinstance(param, LoopList):
            return param[self.loop_idx]
        return param

    def evaluate(self):
        pass

    def evaluate_loop(self, loop):
        pass


class ULParameterNode(ULLogicNode):

    def __init__(self):
        self._fetched = False
        super().__init__()

    @property
    def fetched(self):
        return self._fetched

    @fetched.setter
    def fetched(self, val):
        if val and not self._fetched:
            self.fetch()
        self._fetched = val

    def reset(self):
        self._fetched = False
        return super().reset()

    def fetch(self):
        pass


class ULActionNode(ULLogicNode):

    def __init__(self):
        self._done = False
        self.condition = False
        self.get_condition = self._get_condition
        super().__init__()

    def get_done(self):
        return self._done

    def reset(self):
        super().reset()
        self.get_condition = self._get_condition
        self._done = False

    def _get_condition(self, socket=None):
        condition = socket if socket else self.condition
        if isinstance(condition, ULLogicBase):
            condition = condition.get_value(self)
        if isinstance(condition, LoopList):
            if not self.loop_mode:
                self.loop_mode = True
                self.loop_size = len(condition)
                self.get_condition = self._get_condition_loop
                self.evaluate_loop(condition)
            return False
        return condition

    def _get_condition_loop(self, socket=None):
        condition = socket if socket else self.condition
        if isinstance(condition, ULLogicBase):
            return condition.get_value(self)[self.loop_idx]
        return socket
    
    def evaluate_loop(self, loop):    
        self.loop_idx = 0
        dones = LoopList()
        for condition in loop:
            if condition:
                self.evaluate()
                dones.append(self._done)
            self.loop_idx += 1
        self._done = dones


class ULConditionNode(ULLogicNode):

    def __init__(self):
        self.get_condition = self._get_condition
        super().__init__()

    def _get_condition(self, socket=None):
        condition = socket if socket else self.condition
        if isinstance(condition, ULLogicBase):
            condition = condition.get_value(self)
        if isinstance(condition, LoopList):
            if not self.loop_mode:
                self.loop_mode = True
                self.loop_size = len(condition)
            return condition[self.loop_idx]
        return condition

    def _get_condition_loop(self, socket=None):
        condition = socket if socket else self.condition
        if isinstance(condition, ULLogicBase):
            return condition.get_value(self)[self.loop_idx]
        return socket

    def reset(self):
        super().reset()
        self.get_condition = self._get_condition


class LogicNodeCustom(ULLogicNode):
    pass



from uplogic.nodes import ULActionNode


class ULInstallSubNetwork(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.target_object = None
        self.tree_name = None
        self.initial_status = None
        self._network = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def setup(self, network):
        self._network = network

    def evaluate(self):
        if not self.get_condition():
            return
        target_object = self.get_input(self.target_object)
        tree_name = self.get_input(self.tree_name)
        initial_status = self.get_input(self.initial_status)
        self._network.install_subnetwork(
            target_object,
            tree_name,
            initial_status
        )
        self._done = True



import bge
import bpy
import pickle
import socket
import threading
from uplogic.utils.constants import DISCONNECT_MSG
from uplogic.console import error, success, debug
import time


class Server:

    def __init__(self, ip=None, port=8303, start=False):
        if ip is None:
            ip = socket.gethostbyname(socket.gethostname())
        self.ip = ip
        self.clients = []
        self.port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.scene = bge.logic.getCurrentScene()
        self.running = False
        if start:
            self.start()

    def start(self):
        if self.running:
            print('Server already running.')
            return
        try:
            print('Starting server...')
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            bge.logic.getCurrentScene().onRemove.append(self.shutdown)
            self.socket.setblocking(False)
            self.socket.settimeout(.0001)
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.socket.bind((self.ip, self.port))
            self.running = True
            thread = threading.Thread(target=self.update)
            thread.start()
        except socket.error as e:
            error(e)
            self.shutdown()
            return
        except Exception as e:
            error(e)
            self.shutdown()
            return
        self.socket.listen()
        success('[SUCCESS]')

    def shutdown(self):
        if not self.running:
            print("Server offline.")
            return
        print('Stopping server...')
        scene = bge.logic.getCurrentScene()
        if self.shutdown in scene.onRemove:
            scene.onRemove.remove(self.shutdown)
        try:
            self.running = False
            for conn in self.clients:
                conn.close()
            if self.clients:
                self.socket.shutdown(socket.SHUT_WR)
            self.clients = []
            self.socket.close()
            success('[SUCCESS]')
        except Exception as e:
            self.running = False
            self.clients = []
            self.socket.close()
            error(e)
            error('Runtime Exit.')

    def restart(self):
        self.shutdown()
        self.start()

    def on_receive(self, msg):
        pass

    def send(self, msg, subject=''):
        if self.socket and self.running:
            if subject:
                msg = {
                    'subject': subject,
                    'content': msg
                }
            for conn in self.clients:
                conn.send(pickle.dumps(msg))

    def threaded_client(self, conn, addr):
        connected = True
        self.clients.append(conn)
        while connected and self.running:
            try:
                bmsg = conn.recv(2**14)
                msg = pickle.loads(bmsg)
                if not bmsg or msg == DISCONNECT_MSG:
                    print('Client disconnected.')
                    self.clients.remove(conn)
                    connected = False
                else:
                    self.on_receive(msg)
            except socket.error as e:
                error(f'{e.__class__.__name__}: {e}')
                connected = False
            except pickle.UnpicklingError as e:
                pass
            except Exception as e:
                error(f'{e.__class__.__name__}: {e}')
                error('Threaded Client Error')
                connected = False
        conn.send(pickle.dumps(DISCONNECT_MSG))
        print('Closing Connection...')
        conn.close()
        debug(f'[ACTIVE CONNECTIONS] {len(self.clients)}')
        return

    def update(self):
        while self.running:
            try:
                conn, add = self.socket.accept()
                print(f"Connected to: {add}")
                thread = threading.Thread(target=self.threaded_client, args=(conn, add))
                thread.start()
                debug(f'[ACTIVE CONNECTIONS] {len(self.clients)}') # {threading.active_count() - 2}')
            except BlockingIOError:
                pass
            except socket.timeout:
                pass
            except TimeoutError:
                pass
            except OSError as e:
                self.running = False
                self.shutdown()
            except Exception as e:
                print(e, 'Exception')
                self.running = False
                self.shutdown()
        print('Server Shut Down')



from uplogic.nodes import ULParameterNode


class ULGetCollection(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.collection = None
        self.OUT = self.add_output(self.get_collection)

    def get_collection(self):
        return self.get_input(self.collection)



from bge import logic
from uplogic.nodes import ULParameterNode
from mathutils import Vector


class ULGamepadTrigger(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.index = None
        self.sensitivity = None
        self.threshold = None
        self._values = None
        self.LEFT = self.add_output(self.get_left)
        self.RIGHT = self.add_output(self.get_right)

    @property
    def values(self) -> Vector:
        if self._values is not None:
            return self._values
        index = self.get_input(self.index)
        sensitivity = self.get_input(self.sensitivity)
        if logic.joysticks[index]:
            joystick = logic.joysticks[index]
        else:
            self._values = Vector((0, 0))
            return self._values
        self._values = Vector((
            joystick.axisValues[4] * sensitivity,
            joystick.axisValues[5] * sensitivity
        ))
        return self._values

    def get_left(self):
        threshold = self.get_input(self.threshold)
        value = self.values.x
        if -threshold < value < threshold:
            value = 0
        return value

    def get_right(self):
        threshold = self.get_input(self.threshold)
        value = self.values.y
        if -threshold < value < threshold:
            value = 0
        return value

    def reset(self):
        super().reset()
        self._values = None



from uplogic.nodes import ULParameterNode
from uplogic.serialize import Vec2, Vec3, Vec4, Mat3, Mat4, GameObj


class ULSerializeData(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.data = None
        self.serialize_as = 'builtin'
        self.dattypes = {
            'Vec2': Vec2,
            'Vec3': Vec3,
            'Vec4': Vec4,
            'Mat3': Mat3,
            'Mat4': Mat4,
            'GameObj': GameObj,
        }
        self.OUT = self.add_output(self.get_data)

    def get_data(self):
        dat = self.get_input(self.data)
        if self.serialize_as == 'builtin':
            return dat
        return self.dattypes[self.serialize_as](dat)



from uplogic.nodes import ULActionNode
from uplogic.animation.rig import Rig, RigBone
from bge.logic import ROT_MODE_QUAT


class SetRigBoneAttributeNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.armature = None
        self.bone = None
        self.value = None
        self.attribute = 'inherit_rotation'
        self.scale_mode = 'FULL'
        self._rig: Rig = None
        self.DONE = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        armature = self.get_input(self.armature)
        if self._rig is None or self._rig.game_object is not armature:
            self._rig = Rig(armature)

        bone: RigBone = self._rig.bones[self.get_input(self.bone)]

        value = self.get_input(self.value)
        print(self.attribute, value, bone.name)
        setattr(bone, self.attribute, value if value is not None else self.scale_mode)
        for bone in self._rig.bones.values():
            print(bone.name, bone.inherit_rotation)
        # self._rig.blenderObject.update_tag()
        self._done = True



from uplogic.nodes import ULConditionNode


class ULMouseScrolled(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.wheel_direction = 0
        self.OUT = self.add_output(self.get_wheel)
        self.DIFF = self.add_output(self.get_diff)

    def get_wheel(self):
        wd = self.wheel_direction
        if wd == 1:  # UP
            return self.network.mouse.wheel > 0
        elif wd == 2:  # DOWN
            return self.network.mouse.wheel < 0
        elif wd == 3:  # UP OR DOWN
            return self.network.mouse.wheel != 0

    def get_diff(self):
        return self.network.mouse.wheel



from uplogic.nodes import ULParameterNode


class ULInitNewDict(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.key = None
        self.val = None
        self.DICT = self.add_output(self.get_dict)

    def get_dict(self):
        key = self.get_input(self.key)
        value = self.get_input(self.val)
        return {str(key): value}



from uplogic.nodes import ULConditionNode
from bge.logic import keyboard


class ULKeyboardActive(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return len(keyboard.activeInputs) > 0



from uplogic.nodes import ULActionNode


class ULSetParent(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.child_object = None
        self.parent_object = None
        self.compound = True
        self.ghost = True
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        child_object = self.get_input(self.child_object)
        parent_object = self.get_input(self.parent_object)
        compound = self.get_input(self.compound)
        ghost = self.get_input(self.ghost)
        child_object.setParent(parent_object, compound, ghost)
        self._done = True



from uplogic.nodes import ULParameterNode


class ULDictValue(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.dict = None
        self.key = None
        self.default_value = None
        self.OUT = self.add_output(self.get_val)

    def get_val(self):
        dictionary = self.get_input(self.dict)
        if dictionary is None:
            return None
        key = self.get_input(self.key)
        return dictionary.get(
            key,
            self.get_input(self.default_value)
        )



from .widget import Widget
from .button import Button
from .layout import RelativeLayout
from bge import render
from uplogic.input.mouse import MOUSE
from uplogic.utils.math import clamp, map_range


class Slider(RelativeLayout):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 10.],
        relative={},
        halign='left',
        valign='bottom',
        orientation='horizontal',
        bar_width=.3,
        bar_color=(1, 1, 1, 1),
        bar_hover_color=(1, 1, 1, 1),
        knob_size=1,
        knob_color=(.8, .8, .8, 1),
        knob_hover_color=(.8, .8, .8, 1),
        steps=-1,
        allow_bar_click=True,
        angle=0
    ):
        self.bar = None
        self.knob = None
        super().__init__(pos, size, (0, 0, 0, 0), relative, halign=halign, valign=valign, angle=angle)
        self._value = 0
        # self.border_width = 1
        # self.border_color = (0, 0, 0, 1)
        self._horiz = orientation != 'vertical'
        bar_size = (1, bar_width) if self._horiz else (bar_width, 1)
        self.bar = Button(
            pos=(.5, .5),
            size=bar_size,
            valign='center',
            halign='center',
            relative={'size': True, 'pos': True},
            bg_color=bar_color,
            click_color=bar_color,
            hover_color=bar_hover_color
        )
        self._slide_axis = 0 if self._horiz else 1
        self._up_axis = 1 if self._horiz else 0
        self.add_widget(self.bar)
        if allow_bar_click:
            self.bar.on_hold = self._on_press
        self.knob = Button(size=(0, 0), bg_color=knob_color, click_color=knob_color, hover_color=knob_hover_color, valign='center', halign='center')
        self.knob_size = knob_size
        self.knob.on_hold = self._on_press
        self.add_widget(self.knob)
        self.steps = steps
        self.value = 0
        self._clicked = False
        self._released = False
        self._in_focus = False
        self._down = False
        self.start()

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, val):
        if self.parent is not val and self.parent:
            self.parent.remove_widget(self)
        if self.use_clipping is None:
            self.use_clipping = val.use_clipping
        self._parent = val
        self.pos = self.pos
        self.size = self.size
        self._build_shader()

    def _build_shader(self):
        super()._build_shader()
        if self.bar and self.knob:
            self.bar._build_shader()
            self.knob._build_shader()
        if self.parent:
            self.value = self.value

    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, value):
        newval = value
        if self.steps > 0:
            newval -= newval % (1 / self.steps)
        newval = clamp(newval)
        if newval != self._value:
            self._value = newval
            self.on_value(self)
        knob_size = (self.knob._draw_size[self._slide_axis] * .5)
        self.knob.pos[self._slide_axis] = map_range(self._value, 0, 1, knob_size + self.border_width * .5, self._draw_size[self._slide_axis] - knob_size - self.border_width * .5)
        size = self._draw_size
        self.knob.pos[self._up_axis] = (size[1] * .5) if self._horiz else (size[0] * .5)
        knob_width = (size[1] if self._horiz else size[0])
        self.knob.size = (knob_width * self.knob_size - self.border_width, knob_width * self.knob_size - self.border_width)

    def _on_press(self, widget):
        pos = self._draw_pos
        knob_size = self.knob._draw_size[self._slide_axis] * .5
        step_size = (self._draw_size[self._slide_axis] / self.steps) * .5 if self.steps > 0 else 0
        self.value = map_range(
            (MOUSE.position.x * render.getWindowWidth() + step_size if self._horiz else (1 - MOUSE.position.y) * render.getWindowHeight() + step_size),
            pos[self._slide_axis] + knob_size,
            pos[self._slide_axis] + self._draw_size[self._slide_axis] - knob_size,
            0,
            1
        )

    def on_hold(self, widget):
        pass

    def on_release(self, widget):
        pass

    def on_hover(self, widget):
        pass

    def on_value(self, widget):
        pass


class FrameSlider(Slider):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 10.],
        relative={},
        halign='left',
        valign='bottom',
        orientation='horizontal',
        border_width=1,
        border_color=(.8, .8, .8, 1),
        bg_color=(0, 0, 0, 0),
        bar_color=(1, 1, 1, 1),
        bar_hover_color=(.1, .1, .1, .3),
        steps=-1,
        allow_bar_click=True,
        angle=0
    ):
        self.bar = None
        self.knob = None
        Widget.__init__(self, pos, size, (0, 0, 0, 0), relative, halign=halign, valign=valign, angle=angle)
        self._value = 0
        self._horiz = orientation != 'vertical'
        self.bar = Button(
            pos=(.5, .5),
            size=(1, 1),
            valign='center',
            halign='center',
            relative={'size': True, 'pos': True},
            border_color=border_color,
            bg_color=bg_color,
            click_color=bg_color,
            border_width=border_width,
            hover_color=bar_hover_color
        )
        self.border_color = border_color
        self.border_width = border_width
        self._slide_axis = 0 if self._horiz else 1
        self._up_axis = 1 if self._horiz else 0
        self.knob_size = 1
        if allow_bar_click:
            self.bar.on_hold = self._on_press
        self.knob = RelativeLayout(size=(1, 1), bg_color=bar_color, valign='center', halign='center')
        self.add_widget(self.bar)
        self.add_widget(self.knob)
        self.steps = steps
        self.value = 0
        self._clicked = False
        self._released = False
        self._in_focus = False
        self._down = False
        self.start()


class ProgressSlider(FrameSlider):

    def __init__(self, pos=[0, 0], size=[100, 10], relative={}, halign='left', valign='bottom', orientation='horizontal', border_width=1, border_color=(0.8, 0.8, 0.8, 1), bg_color=(0, 0, 0, 0), bar_color=(1, 1, 1, 1), bar_hover_color=(0.1, 0.1, 0.1, 0.3), steps=-1, allow_bar_click=True, angle=0):
        self.bar = None
        self.knob = None
        super().__init__(pos, size, relative, halign, valign, orientation, border_width, border_color, bg_color, bar_color, bar_hover_color, steps, allow_bar_click, angle)
        self.knob.relative = {'size': True}
        self.knob.halign = 'left'
        self.knob.valign = 'bottom'
        self.start()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        newval = value
        if self.steps > 0:
            newval -= newval % (1 / self.steps)
        newval = clamp(newval)
        if newval != self._value:
            self._value = newval
            self.on_value(self)
        ds = self._draw_size
        width_sub = self.border_width / ds[0]
        height_sub = self.border_width / ds[1]
        self.knob.size = (clamp(self._value - width_sub), 1 - height_sub) if self._horiz else (1 - width_sub, clamp(self._value - height_sub))
        if self.border_width and self.border_color[3] > 0:
            self.knob.y = self.border_width * .5
            self.knob.x = self.border_width * .5

    def _on_press(self, widget):
        pos = self._draw_pos
        step_size = (self._draw_size[self._slide_axis] / self.steps) * .5 if self.steps > 0 else 0
        self.value = map_range(
            (MOUSE.position.x * render.getWindowWidth() + step_size if self._horiz else (1 - MOUSE.position.y) * render.getWindowHeight() + step_size),
            pos[self._slide_axis],
            pos[self._slide_axis] + self._draw_size[self._slide_axis],
            0,
            1
        )



from uplogic.nodes import ULConditionNode


class ULOnNextTick(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.input_condition = None
        self._activated = 0
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        condition = self.get_input(self.input_condition)
        if self._activated == 1:
            if not condition:
                self._activated = 0
            return True
        elif condition:
            self._activated = 1
            return False
        elif self._activated == 0:
            return False



from bge import logic, render
from io import StringIO
from uplogic.ui.canvas import Canvas
from uplogic.ui.layout import RelativeLayout
from uplogic.ui.label import Label
from uplogic.ui.textinput import TextInput
from uplogic.data import GlobalDB
from uplogic.utils.raycasting import raycast_mouse
from uplogic.utils.scene import world_to_screen
from uplogic.utils.math import cycle
from uplogic.utils.math import clamp
from uplogic.input.keyboard import key_down, key_pulse
from uplogic.input.mouse import mouse_down, mouse_wheel
import bpy, blf
import sys, os
from datetime import datetime
from mathutils import Vector
import uplogic


GLOBALS = {}


def set_global(key, val):
    global GLOBALS
    GLOBALS[key] = val


def _get_globals():
    scene = logic.getCurrentScene()
    global GLOBALS
    GLOBALS['scene'] = scene
    GLOBALS['bpy'] = bpy
    GLOBALS['bge'] = bge
    GLOBALS['uplogic'] = uplogic
    GLOBALS['logic'] = logic
    GLOBALS['render'] = render
    GLOBALS['console'] = get_console()
    for obj in scene.objects:
        GLOBALS[obj.blenderObject.name] = obj
    return GLOBALS


def enable(toggle_key='F12', visible=False):
    get_console(True, toggle_key=toggle_key, visible=visible)
    sys.stdout = Console()


def disable():
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
    console = get_console()
    if console:
        console.stop()
    consoles = GlobalDB.retrieve('uplogic.consoles')
    consoles.remove('default')


class Console(StringIO):

    def write(self, __s: str) -> int:
        log(__s)


class ErrorConsole(StringIO):
    def write(self, __s: str) -> int:
        error(__s)

COLORS = {
    'INFO': [1, 1, 1, 1],
    'DEBUG': [1, 1, .6, 1],
    'WARNING': [1, .8, .2, 1],
    'ERROR': [1, .3, .3, 1],
    'SUCCESS': [.3, 1, .3, 1]
}

class ConsoleLayout(Canvas):
    opacity = 1
    padding = [5, 10]
    toggle_key = 'F12'

    def __init__(self, toggle_key='F12', visible=False):
        scene = logic.getCurrentScene()
        self.toggle_key = toggle_key
        self._mouse_visible = False
        self.issued_commands = []
        self._mouse_down = False
        self._goback_index = -1
        super().__init__()
        if not getattr(bpy.context.scene, 'screen_console_open', False) and not visible:
            self.show = False
        self.input = TextInput(text='', shadow=True, valign='center')
        self.input.on_enter = self.on_enter
        self.input.edit = True
        self.layout = RelativeLayout(relative={'size': True, 'pos': True}, pos=[0, 0], size=(1, .4), bg_color=[0, 0, 0, .4])
        self.layout.use_clipping = True
        self.add_widget(self.layout)
        self.layout.add_widget(self.input)
        self.fade_event = None
        self._toggle_key = False
        self._prev_msg = None
        if disable not in scene.onRemove:
            scene.onRemove.append(disable)
        if self.update not in scene.pre_draw:
            scene.pre_draw.append(self.update)
        self.nameplate = Label(text='', shadow=True, relative={'pos': True}, halign='center', font_size=13)
        self.nameplate.update = self.update_nameplate
        self.canvas.add_widget(self.nameplate)
        self.font_size = 14
        self.position = 'bottom'

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, val):
        if val == 'center':
            self.layout.valign = val
            self.layout.halign = val
            self.layout.pos = (.5, .5)
            self.layout.size = (.4, .4)
            self._position = val
        elif val == 'left':
            self.layout.valign = 'bottom'
            self.layout.halign = val
            self.layout.size = (.4, 1)
            self.layout.pos = (0, 0)
            self._position = val
        elif val == 'right':
            self.layout.valign = 'bottom'
            self.layout.halign = val
            self.layout.pos = (1, 0)
            self.layout.size = (.4, 1)
            self._position = val
        elif val == 'top':
            self.layout.valign = 'top'
            self.layout.halign = 'left'
            self.layout.pos = (0, 1)
            self.layout.size = (1, .4)
            self._position = val
        elif val == 'bottom':
            self.layout.valign = 'bottom'
            self.layout.halign = 'left'
            self.layout.pos = (0, 0)
            self.layout.size = (1, .4)
            self._position = val
        else:
            error(f'"{val}" not recognized.')
        self.arrange()

    @property
    def font_size(self):
        return self._font_size

    @font_size.setter
    def font_size(self, val):
        val = clamp(val, 5, 30)
        self._font_size = int(val)
        for c in self.layout.children:
            c.font_size = int(val)
        self.nameplate.font_size = int(val)
        self.arrange()

    def on_enter(self):
        self.issued_commands.append(self.input.text)
        sys.__stdout__.write(f'{self.input.text}\n')
        self.add_message(self.input.text)
        command_name = self.input.text.split(' ')[0]
        command = Commands.commands.get(command_name)
        if command:
            command.invoke(self.input.text)
        else:
            try:
                exec(self.input.text, _get_globals())
            except Exception as e:
                error(f'Error occured when executing command "{command_name}":\n{e}')
        self._goback_index = -1
        self.input.text = ''
        self.input.edit = True

    def update_nameplate(self):
        ray = raycast_mouse()
        mdown = mouse_down()
        if ray.obj:
            self.nameplate.pos = world_to_screen(ray.point) + Vector((0, .01))
            self.nameplate.text = ray.obj.blenderObject.name
            if mdown and not self._mouse_down:
                self.input.write(self.nameplate.text)
        else:
            self.nameplate.text = ''
        self._mouse_down = mdown

    def update(self):
        if self.input.edit != self.show:
            self.input.edit = self.show
        move_goback = key_pulse('UPARROW') - key_pulse('DOWNARROW')
        if key_down(self.toggle_key):
            if not self.show:
                self._mouse_visible = logic.mouse.visible
            if not self._toggle_key:
                self.show = not self.show
                self.opacity = 1
            self._toggle_key = True
        elif not self.show:
            self._toggle_key = False
            return
        elif move_goback:
            if not self.issued_commands:
                pass
            elif not self._toggle_key:
                self._goback_index = clamp(self._goback_index + move_goback, -1, len(self.issued_commands) - 1)
                if self._goback_index < 0:
                    self.input.text = ''
                else:
                    self.input.text = list(self.issued_commands.__reversed__())[self._goback_index]
                self.input.move_cursor_to_end()
            self._toggle_key = True
        else:
            self._toggle_key = False
        logic.mouse.visible = self.show or self._mouse_visible

    def stop(self):
        self.clear()
        scene = logic.getCurrentScene()
        if self.toggle in scene.pre_draw:
            scene.pre_draw.remove(self.toggle)

    def add_message(self, msg, type='INFO', time=True):
        if (msg == ' ' or self._prev_msg == ' ') and len(self.layout.children):
            self.layout.children[-1].text += msg
            self._prev_msg = msg
            return
        now = datetime.now()
        current_time = f'[{now.strftime("%H:%M:%S")}]' if time else "\t\t\t\t  ".replace('\t', '    ')
        self.layout.add_widget(Label(text=f'{current_time}  {msg}', pos=[5, 10], font_color=COLORS[type], shadow=True, font_size=self.font_size))
        self._prev_msg = msg
        self.arrange()

    def arrange(self):
        blf.size(0, self.font_size)
        dim = blf.dimensions(0, 'A')
        cheight = dim[1]
        cwidth = dim[0]
        lheight = self.layout._draw_size[1]
        amount = lheight / cheight
        y = cheight * 2.4 * 1.5
        for i, child in enumerate(self.layout._children_reversed):
            if child is self.input:
                continue
            child.pos[1] = y
            y += cheight * 1.5
            if child.pos[1] > lheight - cheight:
                self.layout.remove_widget(child)
            child.opacity = 1 - (i * (1/amount))
            child.shadow_color[3] = child.font_color[3]
        self.input.pos[1] = self.font_size + 3
        self.input.pos[0] = cwidth


def get_console(create=False, toggle_key='F12', visible=False) -> ConsoleLayout:
    consoles = GlobalDB.retrieve('uplogic.consoles')
    console = consoles.get('default')
    if console is None and create:
        console = ConsoleLayout(toggle_key=toggle_key, visible=visible)
        consoles.put('default', console)
        debug('On-Screen Console active; Check System Console for Errors.')
    return console


class ansicol:
    RED = '\033[31m\033[1m'
    GREEN = '\033[32m\033[1m'
    YELLOW = '\033[33m'
    BYELLOW = '\033[93m'
    BBLUE = '\033[36m'
    END = '\033[0m'


def write(msg, type):
    _f = {
        'INFO': log,
        'DEBUG': debug,
        'WARNING': warning,
        'ERROR': error,
        'SUCCESS': success
    }

    _f[type](msg)


def log(msg, type='INFO'):
    console = get_console()

    if console is None:
        print(msg)
        return
    show_time = True
    for msg in str(msg).split('\n'):
        if msg:
            msg = msg.replace('  ', '    ')
            console.add_message(f'{msg}', type, time=show_time)
            show_time = False
            sys.__stdout__.write(f'{msg}\n')


def warning(msg):
    console = get_console()
    sysmsg = f'{ansicol.YELLOW}Warning{ansicol.END}: {msg}'
    if console is None:
        print(sysmsg)
        return
    show_time = True
    for msg in str(msg).split('\n'):
        if msg:
            msg.replace('  ', '    ')
            console.add_message(f'{msg}', 'WARNING', time=show_time)
            show_time = False
            sys.__stdout__.write(f'{sysmsg}\n')


def error(msg):
    console = get_console()
    sysmsg = f'{ansicol.RED}Error{ansicol.END}: {msg}'
    if console is None:
        print(sysmsg)
        # print(msg)
        return
    show_time = True
    for msg in str(msg).split('\n'):
        if msg:
            msg.replace('  ', '    ')
            console.add_message(f'{msg}', 'ERROR', time=show_time)
            sys.__stdout__.write(f'{sysmsg}\n')
            show_time = False


def success(msg):
    console = get_console()
    sysmsg = f'{ansicol.GREEN}Success{ansicol.END}: {msg}'
    if console is None:
        print(sysmsg)
        return
    show_time = True
    for msg in str(msg).split('\n'):
        if msg:
            msg.replace('  ', '    ')
            console.add_message(f'{msg}', 'SUCCESS', time=show_time)
            sys.__stdout__.write(f'{sysmsg}\n')
            show_time = False


def debug(msg):
    console = get_console()
    sysmsg = f'{ansicol.BYELLOW}Debug{ansicol.END}: {msg}'
    if console is None:
        print(sysmsg)
        return
    show_time = True
    for msg in str(msg).split('\n'):
        if msg:
            msg.replace('  ', '    ')
            console.add_message(f'{msg}', 'DEBUG', time=show_time)
            sys.__stdout__.write(f'{sysmsg}\n')
            show_time = False

nodeprefs = bpy.context.preferences.addons.get('bge_netlogic', None)
if nodeprefs and getattr(bpy.context.scene, 'use_screen_console', True):
    enable(toggle_key='F12')


import bpy
import bge




class Commands:
    commands = {}

    @classmethod
    def add_command(cls, command):
        cls.commands[command.command] = command


# Commands.add_command(RemoveObjectCommand)
# Commands.add_command(DisableCommand)
# Commands.add_command(EnableCommand)
# Commands.add_command(ShowInfoCommand)
# Commands.add_command(QuitCommand)
# Commands.add_command(RestartCommand)
# Commands.add_command(PrintCommand)
# Commands.add_command(PauseCommand)


def add_command(command):
    Commands.add_command(command)


def console_command(command):
    Commands.add_command(command)

class Command:
    command = ''
    usage = ''
    arg_count = 0
    description = ''

    @classmethod
    def invoke(cls, message):
        args = message.split(' ')
        args = args[1:]
        if len(args) < cls.arg_count:
            debug(f'Usage: "{cls.command} {cls.usage}"')
            return
        try:
            cls.execute(args)
        except Exception as e:
            error(e)

    @classmethod
    def execute(cls, args):
        pass


@console_command
class RemoveObjectCommand(Command):
    command = 'remove'
    arg_count = 1
    usage = 'OBJECT_ID'
    description = 'Remove an object from the current scene.'

    @classmethod
    def execute(cls, args):
        object_name = args[0]
        scene = bge.logic.getCurrentScene()
        scene.getGameObjectFromObject(
            bpy.data.objects[object_name]
        ).endObject()


@console_command
class DisableCommand(Command):
    command = 'disable'
    arg_count = 1
    usage = 'OBJECT_ID'
    description = 'Set an object to invisible and inactive.'

    @classmethod
    def execute(cls, args):
        object_name = args[0]
        scene = bge.logic.getCurrentScene()
        scene.getGameObjectFromObject(
            bpy.data.objects[object_name]
        ).setVisible(False)


@console_command
class EnableCommand(Command):
    command = 'enable'
    arg_count = 1
    usage = 'OBJECT_ID'
    description = 'Set an object to visible and active.'

    @classmethod
    def execute(cls, args):
        object_name = args[0]
        scene = bge.logic.getCurrentScene()
        scene.getGameObjectFromObject(
            bpy.data.objects[object_name]
        ).setVisible(True)


@console_command
class ShowInfoCommand(Command):
    command = 'showinfo'
    arg_count = 1
    usage = 'STAGE(0-3)'
    description = 'Show game info. 1: Show FPS, 2: Show Profile, 3: Show Properties'

    @classmethod
    def execute(cls, args):
        stage = args[0]
        render.showFramerate(int(stage) > 0)
        render.showProfile(int(stage) > 1)
        render.showProperties(int(stage) > 2)


@console_command
class QuitCommand(Command):
    command = 'quit'
    arg_count = 0
    usage = ''
    description = 'Quit the game.'

    @classmethod
    def execute(cls, args):
        logic.endGame()


@console_command
class RestartCommand(Command):
    command = 'restart'
    arg_count = 0
    usage = ''
    description = 'Restart the game.'

    @classmethod
    def execute(cls, args):
        logic.restartGame()


@console_command
class PrintCommand(Command):
    command = 'print'
    arg_count = 1
    usage = 'MESSAGE'
    description = 'Print a message or variable.'

    @classmethod
    def execute(cls, args):
        msg = args[0]
        print(eval(msg, _get_globals()))


# @console_command
# class PauseCommand(Command):
#     command = 'pause'
#     arg_count = 1
#     usage = 'STATE(0 or 1)'

#     @classmethod
#     def execute(cls, args):
#         state = args[0]
#         if state:
#             logic.getCurrentScene().suspend()
#         else:
#             logic.getCurrentScene().resume()


@console_command
class HelpCommand(Command):
    command = 'help'
    description = 'Print out all available commands.'

    @classmethod
    def execute(cls, args):
        mode = None
        msg = 'Available Commands:'
        if len(args) > 0:
            mode = args[0]
        for command in Commands.commands.values():
            if mode == '-d':
                msg += f'\n -  "{command.command} {command.usage}"    -    {command.description}'
            else:
                msg += f'\n -  "{command.command} {command.usage}"'
        print(msg)


@console_command
class FontSizeCommand(Command):
    command = 'fontsize'
    arg_count = 1
    usage = 'PX_SIZE'
    description = 'Modify the font size of the console.'

    @classmethod
    def execute(cls, args):
        size = args[0]
        get_console().font_size = int(size)


from uplogic.nodes import ULParameterNode
from os import path, pardir
from ...console import error
import bpy


class GetMasterFolderNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.name = ''
        self.PATH = self.add_output(self._get_path)

    def _get_path(self):
        name = self.get_input(self.name)

        parent_dir = path.join(bpy.path.abspath('//'))[:-1]
        while not parent_dir.endswith(name):
            pdir = path.abspath(path.join(parent_dir, pardir))
            if pdir == parent_dir:
                error("Can't go beyond drive's root!")
                return ''
            parent_dir = pdir
        return parent_dir



from uplogic.nodes import ULActionNode
from uplogic.utils.raycasting import raycast, RayCastData
from uplogic.utils.math import get_bitmask
from bge.logic import getCurrentScene
from bpy.types import Material


class ULRaycast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.origin = None
        self.destination = None
        self.local: bool = None
        self.property_name: str = None
        self.material: str = None
        self.exclude: str = None
        self.xray: bool = None
        self.custom_dist: bool = None
        self.distance: float = None
        self.mask: int = get_bitmask(all=True)
        self.visualize: bool = None
        self._data = RayCastData((None, None, None, None, None, None))
        self.face_data = False
        self.RESULT = self.add_output(self.get_result)
        self.PICKED_OBJECT = self.add_output(self.get_picked_object)
        self.POINT = self.add_output(self.get_point)
        self.NORMAL = self.add_output(self.get_normal)
        self.DIRECTION = self.add_output(self.get_direction)
        self.MATERIAL = self.add_output(self.get_material)
        self.UV = self.add_output(self.get_uv)
        self.network = None

    def setup(self, network):
        self.network = network

    def get_result(self):
        return self._data.obj is not None

    def get_picked_object(self):
        return self._data.obj

    def get_point(self):
        return self._data.point

    def get_normal(self):
        return self._data.normal

    def get_direction(self):
        return self._data.direction

    def get_material(self):
        return self._data.face.material if self._data.face else None

    def get_uv(self):
        return self._data.uv

    def evaluate(self):
        condition = self.get_condition()
        if not condition:
            self._data = RayCastData((None, None, None, None, None, None))
            return
        origin = self.get_input(self.origin)
        destination = self.get_input(self.destination)
        local: bool = self.get_input(self.local)
        property_name: str = self.get_input(self.property_name)
        material: Material = self.get_input(self.material)
        exclude: str = self.get_input(self.exclude)
        xray: bool = self.get_input(self.xray)
        distance: float = self.get_input(self.distance)
        visualize: bool = self.get_input(self.visualize)
        self._data = raycast(
            getCurrentScene().active_camera,
            origin,
            destination,
            distance=distance if self.get_input(self.custom_dist) else 0,
            prop=property_name,
            material=material.name if material else '',
            exclude=exclude,
            xray=xray,
            local=local,
            mask=self.get_input(self.mask),
            visualize=visualize,
            face_data=self.face_data
        )



from uplogic.nodes import ULActionNode
from math import degrees
import bpy


class ULSetUIWidgetAttr(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.widget = None
        self.value = None
        self.widget_attr = 'show'
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        widget = self.get_input(self.widget)
        value = self.get_input(self.value)
        if self.widget_attr == 'angle':
            value = degrees(value)
        elif self.widget_attr == 'texture':
            value = value.name
        elif self.widget_attr == 'font':
            value = value.filepath.replace('\\', '/') if value else 0
        setattr(widget, self.widget_attr, value)
        self._done = True



from uplogic.nodes import ULActionNode


class ULSetSensorValue(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_obj = None
        self.sens_name = None
        self.field = None
        self.value = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_obj = self.get_input(self.game_obj)
        sens_name = self.get_input(self.sens_name)
        # sensor = game_obj.blenderObject.game.sensors.get(sens_name)
        sensor = game_obj.sensors.get(sens_name)
        if not sensor:
            return
        field = self.get_input(self.field)
        value = self.get_input(self.value)
        setattr(sensor, field, value)
        self._done = True



from .shader import Filter2D

glsl = """
uniform sampler2D bgl_RenderedTexture;
in vec4 bgl_TexCoord;
out vec4 fragColor;

uniform float power;

void main(void)
{
    vec2 screen_center = vec2(.5, .5);
    float r = length(bgl_TexCoord.xy - screen_center) * 2;
    vec2 direction = normalize(screen_center - bgl_TexCoord.xy);
    float factor = r*(1-power*(r*r));
    vec2 coords = screen_center - direction * .5 * factor;
    vec4 color = texture(bgl_RenderedTexture, coords);
    if (coords.x <= 0.0 || coords.x >= 1.0 || coords.y <= 0.0 || coords.y >= 1.0) {
        color = vec4(0.0, 0.0, 0.0, 1.0);
    }
    fragColor = color;
}"""



class Lens(Filter2D):

    def __init__(self, power=0.0, idx: int = None) -> None:
        self.uniforms = {'power': float(power)}
        super().__init__(glsl, idx, {'power': self.uniforms})

    @property
    def power(self):
        return self.uniforms['power']

    @power.setter
    def power(self, val):
        self.uniforms['power'] = float(val)



from bge import logic
from uplogic.nodes import ULConditionNode
from uplogic.utils.raycasting import raycast_mouse


class ULMousePressedOn(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self.mouse_button = None
        self.OUT = self.add_output(self.get_changed)

    def get_changed(self):
        mouse_button = self.get_input(self.mouse_button)
        game_object = self.get_input(self.game_object)
        mstat = logic.mouse.inputs[mouse_button]
        if not mstat.activated:
            return False
        mpos = logic.mouse.position
        camera = logic.getCurrentScene().active_camera
        vec = 10 * camera.getScreenVect(*mpos)
        ray_target = camera.worldPosition - vec
        distance = camera.getDistanceTo(game_object) * 2.0
        dat = raycast_mouse(distance=distance)
        return dat.obj == game_object



from uplogic.nodes import ULActionNode
from uplogic.utils import debug


class ULPopDictKey(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.dict = None
        self.key = None
        self.new_dict = None
        self.OUT = self.add_output(self.get_done)
        self.DICT = self.add_output(self.get_dict)
        self.VALUE = self.add_output(self.get_value)

    def get_done(self):
        return self._done

    def get_dict(self):
        return self.new_dict

    def get_value(self):
        return self.value

    def evaluate(self):
        if not self.get_condition():
            return
        dictionary = self.get_input(self.dict)
        key = self.get_input(self.key)
        if key in dictionary:
            self.value = dictionary.pop(key)
        else:
            debug("Dict Delete Key Node: Key '{}' not in Dict!".format(key))
            return
        self.new_dict = dictionary
        self._done = True



from uplogic.nodes import ULParameterNode


class ULGetLightEnergy(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.lamp = None
        self.energy = 0
        self.ENERGY = self.add_output(self.get_energy)

    def get_energy(self):
        lamp = self.get_input(self.lamp)
        light = lamp.blenderObject.data
        return light.energy



from .widget import Widget
import gpu
import bpy
from math import ceil
from .widget import rotate2d
from gpu_extras.batch import batch_for_shader
from mathutils import Vector
from os.path import isfile


class Image(Widget):

    vertex_shader = """
    in vec2 texCoord;
    in vec2 pos;
    out vec2 uv;

    uniform mat4 ModelViewProjectionMatrix;

    void main() {
        uv = texCoord;
        gl_Position = ModelViewProjectionMatrix * vec4(pos.xy, 0.0, 1.0);
    }
    """

    fragment_shader = """
    in vec2 uv;
    out vec4 fragColor;

    uniform sampler2D image;
    uniform float alpha = 1.0;

    void main() {
        vec4 color = mix(vec4(0.0), texture(image, uv), alpha);
        fragColor = pow(color, vec4(.5));
    }
    """

    def __init__(self, pos=[0, 0], size=(100, 100), relative={}, texture=None, halign='left', valign='bottom', use_aspect_ratio: bool = True, angle=0):
        self._texture = None
        self._image = None
        self.use_aspect_ratio = use_aspect_ratio
        self._opacity = 1
        super().__init__(pos, size, relative=relative, halign=halign, valign=valign, angle=angle)
        if texture is not None and texture not in bpy.data.images and isfile(texture):
            bpy.data.images.load(texture)
        self.texture = texture
        self.start()

    @property
    def texture(self):
        return self._texture

    @property
    def aspect_ratio(self):
        if self._image is None:
            return 1
        return self._image.size[1] / self._image.size[0]

    @property
    def _draw_size(self):
        size = self.size.copy()
        use_aspect_ratio = self.use_aspect_ratio
        if self.parent is None:
            return size
        if use_aspect_ratio and self._image:
            size[1] = size[0] * self.aspect_ratio
        if self.relative.get('size'):
            pdsize = self.parent._draw_size
            size = [
                size[0] * pdsize[0],
                size[1] * (pdsize[0] if use_aspect_ratio else pdsize[1])
            ]
        return size

    @texture.setter
    def texture(self, val):
        if val is None:
            return
        texture = bpy.data.images.get(val, None)
        if not texture:
            texture = bpy.data.images.load(val)
        self._image = texture
        self.size = self.size
        self._texture = gpu.texture.from_image(texture)

    @property
    def pivot(self):
        """Rotation point for this widget."""
        if self.parent is None:
            return (0, 0)
        v = self._vertices
        x0 = Vector(v[1])
        x1 = Vector(v[0])
        y1 = Vector(v[2])
        y0 = Vector(v[3])
        return self._get_pivot(x0, x1, y0, y1)

    def _build_shader(self):
        pos = self._draw_pos
        size = self._draw_size
        x0 = Vector((pos[0], pos[1]))
        x1 = Vector((pos[0] + size[0], pos[1]))
        y1 = Vector((pos[0] + size[0], pos[1] + size[1]))
        y0 = Vector((pos[0], pos[1] + size[1]))
        pivot = self._get_pivot(x0, x1, y0, y1)

        if self._draw_angle and self._vertices is not None:
            x0 = rotate2d(x0, pivot, self._draw_angle)
            x1 = rotate2d(x1, pivot, self._draw_angle)
            y0 = rotate2d(y0, pivot, self._draw_angle)
            y1 = rotate2d(y1, pivot, self._draw_angle)
        vertices = self._vertices = (
            x1, x0, y1, y0
        )

        self._shader = gpu.types.GPUShader(self.vertex_shader, self.fragment_shader)
        self._batch = batch_for_shader(
            self._shader, 'TRI_STRIP',
            {
                "pos": vertices,
                "texCoord": (
                    (.9999, .0001),
                    (0.0001, 0.0001),
                    (.9999, .9999),
                    (.0001, .9999)
                ),
            },
        )
    
    def draw(self):
        gpu.state.blend_set("ALPHA")
        super()._setup_draw()
        if self.texture is None:
            super().draw()
            return
        self._shader.bind()
        self._shader.uniform_float("alpha", self.opacity)
        self._shader.uniform_sampler("image", self.texture)
        self._batch.draw(self._shader)
        super().draw()


class Sprite(Image):

    tex_vert_shader = """
    in vec2 texCoord;
    in vec2 pos;
    out vec2 uv;

    uniform mat4 ModelViewProjectionMatrix;

    void main() {
        uv = texCoord;
        gl_Position = ModelViewProjectionMatrix * vec4(pos.xy, 0.0, 1.0);
    }
    """

    tex_frag_shader = """
    in vec2 uv;
    out vec4 fragColor;

    uniform sampler2D image;
    uniform float alpha = 1.0;

    void main() {
        vec4 color = mix(vec4(0.0), texture(image, uv), alpha);
        fragColor = pow(color, vec4(.5));
    }
    """

    def __init__(self, pos=[0, 0], size=(100, 100), relative={}, texture=None, idx=0, rows=1, cols=1, halign='left', valign='bottom', use_aspect_ratio=True):
        self._idx = idx
        self.rows = rows
        self.cols = cols
        super().__init__(pos, size, relative, texture, halign=halign, valign=valign, use_aspect_ratio=use_aspect_ratio)

    @property
    def idx(self):
        return self._idx

    @idx.setter
    def idx(self, val):
        self._idx = val
        self._build_shader()

    @property
    def aspect_ratio(self):
        if self._image is None:
            return 1
        return (self._image.size[1] / self.rows) / (self._image.size[0] / self.cols)

    @property
    def rows(self):
        return self._rows

    @rows.setter
    def rows(self, val):
        if val < 1:
            return
        self._rows = val
        self._row_height = 1 / val

    @property
    def cols(self):
        return self._cols

    @cols.setter
    def cols(self, val):
        if val < 1:
            return
        self._cols = val
        self._col_width = 1 / val

    def _build_shader(self):
        pos = self._draw_pos
        size = self._draw_size
        x0 = Vector((pos[0], pos[1]))
        x1 = Vector((pos[0] + size[0], pos[1]))
        y1 = Vector((pos[0] + size[0], pos[1] + size[1]))
        y0 = Vector((pos[0], pos[1] + size[1]))
        pivot = self._get_pivot(x0, x1, y0, y1)

        if self._draw_angle and self._vertices is not None:
            x0 = rotate2d(x0, pivot, self._draw_angle)
            x1 = rotate2d(x1, pivot, self._draw_angle)
            y0 = rotate2d(y0, pivot, self._draw_angle)
            y1 = rotate2d(y1, pivot, self._draw_angle)

        vertices = self._vertices = (
            x1,
            x0,
            y1,
            y0
        )
        self._shader = gpu.types.GPUShader(self.tex_vert_shader, self.tex_frag_shader)
        self._shader.uniform_float("alpha", self.opacity)

        idx = self.idx
        col = idx % self.cols
        col_end = col + 1
        row = ceil((idx + 1) / self.cols) - 1
        row_end = row + 1
        texcoord = (
            (col_end * self._col_width, 1 - row_end * self._row_height),
            (col * self._col_width, 1 - row_end * self._row_height),
            (col_end * self._col_width, 1 - row * self._row_height),
            (col * self._col_width, 1 - row * self._row_height)
        )
        self._batch = batch_for_shader(
            self._shader, 'TRI_STRIP',
            {
                "pos": vertices,
                "texCoord": texcoord
            },
        )


from .widget import Widget
from bge import render
import gpu
import bge


class Canvas(Widget):
    """The base class for UI layouts. This class has no visual representation
    and spans the whole screen. It is intended to manage collections of widgets
    more easily.

    A canvas cannot be attached to another widget and has its own update cycle.
    """

    def __init__(self):
        super().__init__((0, 0), (0, 0))
        self._hover_consumed = False
        self._click_consumed = False
        self._old_width = bge.render.getWindowWidth()
        self._old_height = bge.render.getWindowHeight()
        self.use_clipping = False
        self._to_evaluate: list[Widget] = []
        self.register()
        self.start()
    
    def register(self):
        bge.logic.getCurrentScene().post_draw.append(self.draw)

    def unregister(self):
        while self.draw in bge.logic.getCurrentScene().post_draw:
            bge.logic.getCurrentScene().post_draw.remove(self.draw)

    def remove(self):
        self.unregister()

    def fetch_size(self):
        for c in self.children:
            c.size = c.size
            c.pos = c.pos
            c.parent = c.parent

    @property
    def _draw_pos(self):
        return [0, 0]

    @property
    def pivot(self):
        return (0, 0)

    @property
    def _draw_size(self):
        return [
            render.getWindowWidth(),
            render.getWindowHeight()
        ]

    @property
    def clipping(self):
        return [
            0,
            render.getWindowWidth(),
            render.getWindowHeight(),
            0
        ]

    @property
    def parent(self):
        return None

    @parent.setter
    def parent(self, val):
        self._parent = None

    def draw(self):
        width = bge.render.getWindowWidth()
        height = bge.render.getWindowHeight()
        if width != self._old_width or height != self._old_height:
            self.fetch_size()
        self._old_width = width
        self._old_height = height
        if not self.show:
            return
        gpu.state.blend_set('ALPHA')
        super().draw()
        self._hover_consumed = False
        self._click_consumed = False
        for w in self._to_evaluate.__reversed__():
            w.evaluate()
        while self._to_evaluate:
            self._to_evaluate.pop().update()

    def update(self):
        pass

    def new_layer(self):
        layer = Layer()
        self.add_widget(layer)
        return layer


class Layer(Widget):

    @property
    def _draw_pos(self):
        return [0, 0]

    @property
    def pivot(self):
        return (0, 0)

    @property
    def _draw_size(self):
        return [
            render.getWindowWidth(),
            render.getWindowHeight()
        ]

    @property
    def clipping(self):
        return [
            0,
            render.getWindowWidth(),
            render.getWindowHeight(),
            0
        ]

    @property
    def parent(self):
        """The widget whose position and size to use relatively."""
        return self._parent

    @parent.setter
    def parent(self, val):
        if not isinstance(val, Canvas):
            raise TypeError('ui.Layout can only be added to ui.Canvas type!')
        if self.parent is not val and self.parent:
            self.parent.remove_widget(self)
        if self.use_clipping is None:
            self.use_clipping = val.use_clipping
        self._parent = val
        self.pos = self.pos
        self.size = self.size
        for c in self.children:
            c.parent = c.parent
        self._build_shader()



from uplogic.nodes import ULParameterNode
from uplogic.utils.portals import Portals


class SetPortalNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self._portal = None
    
    @property
    def portal(self):
        return self._portal

    @portal.setter
    def portal(self, val):
        self._portal = Portals.get(val)
        self._portal.node = self



from bge import logic
from bge.types import KX_GameObject
import bpy
from bpy.types import Material, Object
from .errors import LogicControllerNotSupportedError
from .constants import FRONT_AXIS_VECTOR_SIGNED
from .math import project_vector3
from .math import clamp
from .math import get_local
from .math import rotate2d
from .math import rotate3d
from .math import rotate_by_axis
from ..events import schedule
from mathutils import Vector, Matrix, Euler
from math import degrees
from math import radians


def xrot_to(
    rotating_object,
    target_pos,
    front_axis_code=1,
    factor=1
):
    local = get_local(rotating_object, target_pos)
    front = Vector((1, 0)) if front_axis_code in [1, 4] else Vector((0, 1))
    if front_axis_code > 2:
        front.negate()
    angle = Vector((local.y, local.z))
    if angle.length < .001:
        return
    angle = angle.angle_signed(front)
    rotating_object.applyRotation((angle*factor, 0, 0), True)


def yrot_to(
    rotating_object,
    target_pos,
    front_axis_code=1,
    factor=1
):
    local = get_local(rotating_object, target_pos)
    front = Vector((1, 0)) if front_axis_code in [0, 3] else Vector((0, 1))
    if front_axis_code > 2:
        front.negate()
    angle = Vector((local.x, local.z))
    if angle.length < .001:
        return
    angle = angle.angle_signed(front)
    rotating_object.applyRotation((0, angle*factor, 0), True)


def zrot_to(
    rotating_object,
    target_pos,
    front_axis_code=1,
    factor=1
):
    local = get_local(rotating_object, target_pos)
    front = Vector((1, 0)) if front_axis_code in [0, 3] else Vector((0, 1))
    if front_axis_code > 2:
        front.negate()
    angle = Vector((local.x, local.y))
    if angle.length < .001:
        return
    angle = angle.angle_signed(front)
    rotating_object.applyRotation((0, 0, angle*factor), True)


def rotate_to(
    object: KX_GameObject,
    target: Vector,
    rotation_axis: int = 2,
    front_axis: int = 1,
    factor:float = 1
):
    """Rotate an object around a local axis towards a point"""
    front = front_axis
    if front > 2:
        front -= 3
    if rotation_axis == front:
        return
    if rotation_axis == 0:
        xrot_to(
            object,
            target,
            front_axis,
            factor
        )
    elif rotation_axis == 1:
        yrot_to(
            object,
            target,
            front_axis,
            factor
        )
    elif rotation_axis == 2:
        zrot_to(
            object,
            target,
            front_axis,
            factor
        )


def move_to(game_object: KX_GameObject, target: Vector, speed: float):
    direction = (game_object.worldPosition - target)
    if direction.length < speed:
        game_object.worldPosition = target
        return True
    direction.normalize()
    game_object.worldPosition -= direction * speed


def _move_to(
    moving_object,
    destination_point,
    speed,
    time_per_frame,
    dynamic,
    distance
):
    if dynamic:
        direction = (
            destination_point -
            moving_object.worldPosition)
        dst = direction.length
        if(dst <= distance):
            moving_object.worldPosition = destination_point
            return True
        direction.z = 0
        direction.normalize()
        velocity = direction * speed
        velocity.z = moving_object.worldLinearVelocity.z
        moving_object.worldLinearVelocity = velocity
        return False
    else:
        direction = (
            destination_point -
            moving_object.worldPosition
            )
        dst = direction.length
        if(dst <= distance):
            moving_object.worldPosition = destination_point
            return True
        direction.normalize()
        displacement = speed * time_per_frame
        motion = direction * displacement
        moving_object.worldPosition += motion
        return False


def controller_brick_status(owner, controller_name):
    cont = owner.controllers[controller_name]
    state = (
        owner
        .blenderObject
        .game
        .controllers[controller_name]
        .type
    )
    if not cont.sensors:
        return False
    elif state == 'LOGIC_AND':
        return False not in [sens.positive for sens in cont.sensors]
    elif state == 'LOGIC_OR':
        return True in [sens.positive for sens in cont.sensors]
    elif state == 'LOGIC_NAND':
        return False in [sens.positive for sens in cont.sensors]
    elif state == 'LOGIC_NOR':
        return True not in [sens.positive for sens in cont.sensors]
    elif state == 'LOGIC_XOR':
        return [
            sens.positive
            for sens in
            cont.sensors
        ].count(True) % 2 != 0
    elif state == 'LOGIC_XNOR':
        check = cont.sensors[0].positive
        return False not in [
            sens.positive == check
            for sens in
            cont.sensors
        ]
    else:
        raise LogicControllerNotSupportedError


class ControllerBrick(tuple):

    @property
    def brick(self):
        return self[0]

    @property
    def name(self):
        return self[0].name

    @property
    def positive(self):
        return self[1]

    @property
    def sensors(self):
        return self[2]

    @property
    def actuators(self):
        return self[3]


def controller_brick(owner, controller_name):
    cont = owner.controllers[controller_name]
    state = (
        owner
        .blenderObject
        .game
        .controllers[controller_name]
        .type
    )
    if not cont.sensors:
        return ControllerBrick([cont, False, cont.sensors, cont.actuators])
    elif state == 'LOGIC_AND':
        return ControllerBrick([cont, False not in [sens.positive for sens in cont.sensors], cont.sensors, cont.actuators])
    elif state == 'LOGIC_OR':
        return ControllerBrick([cont, True in [sens.positive for sens in cont.sensors], cont.sensors, cont.actuators])
    elif state == 'LOGIC_NAND':
        return ControllerBrick([cont, False in [sens.positive for sens in cont.sensors], cont.sensors, cont.actuators])
    elif state == 'LOGIC_NOR':
        return ControllerBrick([cont, True not in [sens.positive for sens in cont.sensors], cont.sensors, cont.actuators])
    elif state == 'LOGIC_XOR':
        return ControllerBrick([cont, [
            sens.positive
            for sens in
            cont.sensors
        ].count(True) % 2 != 0, cont.sensors, cont.actuators])
    elif state == 'LOGIC_XNOR':
        check = cont.sensors[0].positive
        return ControllerBrick([cont, False not in [
            sens.positive == check
            for sens in
            cont.sensors
        ], cont.sensors, cont.actuators])
    else:
        raise LogicControllerNotSupportedError


def create_curve(
    name: str,
    bevel_depth: float = 0.0,
    dimensions: int = 3,
    material: str or Material = None,
    collection: str = None
) -> KX_GameObject:
    """Create a `KX_GameObject` containing a `bpy.types.Curve` object.

    :param `name`: Name of the new `KX_GameObject`.
    :param `bevel_depth`: Define the "thickness" of the curve. This will add
    geometry along the spline.
    :param `dimensions`: Set the coordinate space in which to calculate the
    curve.
    :param `material`: The material to use for bevel geometry.
    :param `collection`: The collection to which to add the curve. Leave at
    `None` to use scene collection.
    """
    bcurve = bpy.data.curves.new(name, 'CURVE')
    bcurve.bevel_depth = bevel_depth
    bcurve.dimensions = f'{dimensions}D'
    bobj = bpy.data.objects.new(name, bcurve)
    if material:
        if isinstance(material, str):
            bobj.data.materials.append(bpy.data.materials[material])
        elif isinstance(material, Material):
            bobj.data.materials.append(material)
    if collection:
        if isinstance(collection, str):
            collection = bpy.data.collections.get(collection, bpy.context.scene.collection)
    elif collection is None:
        collection = bpy.context.scene.collection
    collection.objects.link(bobj)
    game_obj = logic.getCurrentScene().convertBlenderObject(bobj)
    return game_obj


def set_curve_points(
    curve: KX_GameObject,
    points: list,
    loop: bool = False,
    type: str = 'POLY'
) -> None:
    """Set the curve points of a `KX_GameObject` containing a `bpy.types.Curve` object.

    :param `curve`: `KX_GameObject`
    :param `points`: A list of points to use for the curve.
    """
    bcurve = curve.blenderObject.data
    for spline in bcurve.splines:
        bcurve.splines.remove(spline)
    spline = bcurve.splines.new(type)
    spline.use_cyclic_u = loop
    pos = curve.worldPosition

    spline.points.add(len(points)-1)
    for p, new_co in zip(spline.points, points):
        p.co = ([
            new_co[0] - pos.x,
            new_co[1] - pos.y,
            new_co[2] - pos.z
        ] + [1.0])


class GameObject:

    def __init__(self, game_object: KX_GameObject) -> None:
        self.game_object: KX_GameObject = game_object

    @property
    def blenderObject(self) -> Object:
        return self.game_object.blenderObject

    @property
    def parent(self) -> KX_GameObject:
        return self.game_object.parent

    @parent.setter
    def parent(self, val: KX_GameObject):
        self.game_object.setParent(val)

    def set_parent(self, parent):
        self.game_object.setParent(parent)

    @property
    def children(self):
        return self.game_object.children

    @property
    def children_recursive(self):
        return self.game_object.childrenRecursive

    @property
    def mass(self):
        return getattr(self.game_object, 'mass', 0)

    @mass.setter
    def mass(self, val):
        self.game_object.mass = val

    @property
    def worldPosition(self) -> Vector:
        return self.game_object.worldPosition

    @worldPosition.setter
    def worldPosition(self, val: Vector):
        self.game_object.worldPosition = val

    @property
    def localPosition(self) -> Vector:
        return self.game_object.localPosition

    @localPosition.setter
    def localPosition(self, val: Vector):
        self.game_object.localPosition = val

    @property
    def worldOrientation(self) -> Matrix:
        return self.game_object.worldOrientation

    @worldOrientation.setter
    def worldOrientation(self, val: Matrix):
        self.game_object.worldOrientation = val

    @property
    def localOrientation(self) -> Matrix:
        return self.game_object.localOrientation

    @localOrientation.setter
    def localOrientation(self, val: Matrix):
        self.game_object.localOrientation = val

    @property
    def worldScale(self) -> Vector:
        return self.game_object.worldScale

    @worldScale.setter
    def worldScale(self, val: Vector):
        self.game_object.worldScale = val

    @property
    def localScale(self) -> Vector:
        return self.game_object.localScale

    @localScale.setter
    def localScale(self, val: Vector):
        self.game_object.localScale = val

    @property
    def worldLinearVelocity(self) -> Vector:
        return self.game_object.worldLinearVelocity

    @worldLinearVelocity.setter
    def worldLinearVelocity(self, val: Vector):
        self.game_object.worldLinearVelocity = val

    @property
    def localLinearVelocity(self) -> Vector:
        return self.game_object.localLinearVelocity

    @localLinearVelocity.setter
    def localLinearVelocity(self, val: Vector):
        self.game_object.localLinearVelocity = val

    @property
    def worldAngularVelocity(self) -> Vector:
        return self.game_object.worldAngularVelocity

    @worldAngularVelocity.setter
    def worldAngularVelocity(self, val: Vector):
        self.game_object.worldAngularVelocity = val

    @property
    def localAngularVelocity(self) -> Vector:
        return self.game_object.localAngularVelocity

    @localAngularVelocity.setter
    def localAngularVelocity(self, val: Vector):
        self.game_object.localAngularVelocity = val

    @property
    def worldTransform(self) -> Matrix:
        return self.game_object.worldTransform

    @worldTransform.setter
    def worldTransform(self, val: Matrix):
        self.game_object.worldTransform = val
    
    def move_to(self, target, speed):
        return move_to(self, target, speed)


def get_curve_length(curve: KX_GameObject):
    depsgraph = bpy.context.evaluated_depsgraph_get()
    return sum(s.calc_length() for s in curve.blenderObject.evaluated_get(depsgraph).data.splines)


class Curve(GameObject):
    """Wrapper class for creating and handling curves more easily.

    :param `name`: Name of this curve object.
    :param `bevel_depth`: Define the "thickness" of the curve. This will add
    geometry along the spline.
    :param `dimensions`: Set the coordinate space in which to calculate the
    curve.
    :param `material`: The material to use for bevel geometry.
    :param `collection`: The collection to which to add the curve. Leave at
    `None` to use scene collection.
    """

    _deprecated = False

    def __init__(
        self,
        name: str,
        bevel_depth: float = 0.0,
        dimensions: int = 3,
        material: str or Material =None,
        collection: str = None,
        loop: bool = False,
        type: str = 'POLY'
    ) -> None:
        if self._deprecated:
            print('[UPLOGIC] ULCurve class will be renamed to "Curve" in future releases!')
        self.type = type
        self._loop = loop
        if isinstance(name, KX_GameObject):
            self.game_object = name
            return
        self.game_object = create_curve(
            name=name,
            bevel_depth=bevel_depth,
            dimensions=dimensions,
            material=material,
            collection=collection
        )

    @property
    def name(self):
        """Name of the game object (Read-Only)."""
        return self.game_object.name

    @name.setter
    def name(self, val: str):
        print('Curve.name is Read-Only!')

    @property
    def loop(self):
        """Name of the game object (Read-Only)."""
        return self._loop

    @loop.setter
    def loop(self, val: bool):
        self._loop = val
        self.points = self.points

    @property
    def points(self):
        """Points of the curve in global space."""
        splines = self.game_object.blenderObject.data.splines
        return splines[0].points if len(splines) > 0 else []

    @points.setter
    def points(self, val: list):
        if val != self.points:
            set_curve_points(self.game_object, val, loop=self.loop, type=self.type)

    @property
    def bevel_depth(self):
        """Thickness of the curve geometry as diameter."""
        return self.game_object.blenderObject.data.bevel_depth

    @bevel_depth.setter
    def bevel_depth(self, val):
        self.game_object.blenderObject.data.bevel_depth = val

    @property
    def length(self):
        '''Length of the curve.'''
        depsgraph = bpy.context.evaluated_depsgraph_get()
        return sum(s.calc_length() for s in self.blenderObject.evaluated_get(depsgraph).data.splines)

    @length.setter
    def length(self, val):
        print('Curve.length is read-only!')

    @property
    def path_duration(self):
        '''The number of frames that are needed to traverse the path, defining the maximum value for the "Evaluation Time" setting'''
        return self.blenderObject.data.path_duration

    @path_duration.setter
    def path_duration(self, val):
        self.game_object.blenderObject.data.path_duration = val


class ULCurve(Curve):
    _deprecated = True


class Mesh():

    def __init__(self, mesh: bpy.types.Mesh):
        self.blenderMesh: bpy.types.Mesh = mesh

    def applyRotation(self, rotation, local=False):
        rot = Euler(rotation)
        self.blenderMesh.transform(
            rot.to_matrix().to_4x4()
        )


def add_object(name: str | KX_GameObject, ref: str | KX_GameObject = None, time = 0, dupli = False):
    scene = logic.getCurrentScene()
    if isinstance(name, KX_GameObject):
        name = name.name
    obj = bpy.data.objects.get(name, None)
    if obj is None:
        return
    new_obj = obj.copy()
    new_obj.hide_viewport = False
    if dupli:
        new_obj.data = obj.data.copy()
    bpy.context.scene.collection.objects.link(new_obj)
    game_object = scene.convertBlenderObject(new_obj)
    if ref:
        if isinstance(ref, KX_GameObject):
            ref = ref.name
        ref_obj = scene.objects.get(ref, ref)
        game_object.worldTransform = ref_obj.worldTransform
    if time > 0:
        schedule(game_object.endObject, time)
    return game_object


def add_object_copy(name: str | KX_GameObject, position=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)), scale=Vector((1, 1, 1))):
    orig_ob = bpy.data.objects.get(name, name)
    if orig_ob is None:
        return
    game_scene = logic.getCurrentScene()
    scene = bpy.data.scenes[game_scene.name]
    ob = orig_ob.copy()
    scene.collection.objects.link(ob)
    gobj = game_scene.convertBlenderObject(ob)
    gobj.worldPosition = position
    gobj.worldOrientation = rotation
    gobj.worldScale = scale
    return gobj


def add_object_from_mesh(name: str | KX_GameObject, position=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)), scale=Vector((1, 1, 1))):
    orig_ob = bpy.data.objects.get(name, None)
    if orig_ob is None:
        return
    game_scene = logic.getCurrentScene()
    scene = bpy.data.scenes[game_scene.name]
    ob = bpy.data.objects.new(name, orig_ob.data)
    scene.collection.objects.link(ob)
    gobj = game_scene.convertBlenderObject(ob)
    gobj.worldPosition = position
    gobj.worldOrientation = rotation
    gobj.worldScale = scale
    return gobj



from uplogic.nodes import ULConditionNode
from bge.logic import keyboard


class ULKeyReleased(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.key_code = None
        self.network = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        keycode = self.get_input(self.key_code)
        keystat = keyboard.activeInputs[keycode]
        if self.pulse:
            return (
                keystat.released or
                keystat.inactive
            )
        return keystat.released



from uplogic.nodes import ULActionNode

class ULSendNetworkMessage(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.entity = None
        self.data = None
        self.subject = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        entity = self.get_input(self.entity)
        if entity:
            entity.send(self.get_input(self.data), self.get_input(self.subject))
        self._done = True



from uplogic.nodes import ULConditionNode


class ULLogicTreeStatus(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self.tree_name = None
        self._running = False
        self._stopped = False
        self.tree = None
        self.IFRUNNING = self.add_output(self.get_running)
        self.IFSTOPPED = self.add_output(self.get_stopped)

    def get_running(self):
        tree = self.tree
        if not tree:
            return False
        return tree.is_running()

    def get_stopped(self):
        tree = self.tree
        if not tree:
            return False
        return tree.is_stopped()

    def evaluate(self):
        game_object = self.get_input(self.game_object)
        tree_name = self.get_input(self.tree_name)
        self._running = False
        self._stopped = False
        self.tree = game_object.get(f'IGNLTree_{tree_name}')



from .shader import Filter2D


glsl = """
uniform sampler2D bgl_RenderedTexture;

in vec4 bgl_TexCoord;
uniform float power;
uniform float factor;

out vec4 fragColor;

float map_range(float value, float in_min, float in_max, float out_min, float out_max){
    return (value - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

void main()
{
    float transitionArea = .1f;
    float pow = clamp(power, 0.0, 1.0);
	vec4 px =  texture(bgl_RenderedTexture, bgl_TexCoord.xy);

    float bend_factor = 1.0f;

    float y_offset = abs(bgl_TexCoord.x - 0.5) * bend_factor;
    y_offset *= y_offset;
    y_offset *= abs(bgl_TexCoord.y - 0.5);

    float max_y_offset = .5 * .5 * bend_factor;

    float fac_top = clamp(map_range(bgl_TexCoord.y + y_offset, 1.0-factor + transitionArea + max_y_offset, 1.0-factor + transitionArea, 1, 0), 0, 1);
    float fac_bottom = clamp(map_range(bgl_TexCoord.y - y_offset, factor - transitionArea - max_y_offset, factor - transitionArea, 1, 0), 0, 1);

    fragColor = mix(
        px,
        vec4(0.0, 0.0, 0.0, 1.0),
        fac_bottom + fac_top
    );
}
"""


class Eyelids(Filter2D):

    def __init__(self, power: float = 1.0, factor: float = 0.1, idx: int = None) -> None:
        self.uniforms = {'power': float(power), 'factor': float(factor)}
        super().__init__(glsl, idx, {'power': self.uniforms, 'factor': self.uniforms})

    @property
    def power(self):
        return self.uniforms['power']

    @power.setter
    def power(self, val):
        self.uniforms['power'] = float(val)
    
    @property
    def factor(self):
        return self.uniforms['factor']

    @factor.setter
    def factor(self, val):
        self.uniforms['factor'] = float(val)



from uplogic.nodes import ULConditionNode


class ULHasProperty(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self.property_name = None
        self.mode = 0
        self.OUT = self.add_output(self.get_stat)

    def get_stat(self):
        game_object = self.get_input(self.game_object)
        property_name = self.get_input(self.property_name)
        return (
            property_name in [p[0] for p in game_object.blenderObject.items()]
            if self.mode else
            property_name in game_object.getPropertyNames()
        )



from uplogic.nodes import ULActionNode
import bpy


class ULRunPython(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.module_name = None
        self.module_func = None
        self.mode = 1
        self.arg = None
        self.val = None
        self.OUT = self.add_output(self.get_done)
        self.VAL = self.add_output(self.get_val)
        self._old_mod_name = None
        self._code_as_str = ''
        self._old_mod_fun = None
        self._module = None
        self._modfun = None

    def get_done(self):
        return self._done

    def get_val(self):
        return self.val

    def evaluate(self):
        if not self.get_condition():
            return
        text = self.get_input(self.module_name)
        mfun = self.get_input(self.module_func)
        mname = text.name
        if self.mode:
            mname = mname.split('.')[0]
            if mname and (self._old_mod_name != mname):
                exec("import {}".format(mname))
                self._old_mod_name = mname
                self._module = eval(mname)
            if self._old_mod_fun != mfun:
                self._modfun = getattr(self._module, mfun)
                self._old_mod_fun = mfun
            args = [
                self.get_input(arg)
                for arg in self.arg
            ]
            if args:
                self.val = self._modfun(*args)
            else:
                self.val = self._modfun()
        else:
            if self._old_mod_name != mname:
                self._old_mod_name = mname
                self._code_as_str = compile(bpy.data.texts[mname].as_string(), 'custom', 'exec')
            exec(self._code_as_str, globals())
        self._done = True



from uplogic.nodes import ULActionNode


class ULSetListIndex(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.items: list = None
        self.index: int = None
        self.val = None
        self.new_list: list = None
        self.OUT = self.add_output(self.get_done)
        self.LIST = self.add_output(self.get_list)

    def get_done(self):
        return self._done

    def get_list(self):
        return self.new_list

    def evaluate(self):
        if not self.get_condition():
            return
        list_d: list = self.get_input(self.items)
        index: int = self.get_input(self.index)
        val = self.get_input(self.val)
        list_d[index] = val
        self.new_list = list_d
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic.ui import Canvas


class ULCreateUICanvas(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self._canvas = None
        self.on_init = False
        self.OUT = self.add_output(self._get_done)
        self.CANVAS = self.add_output(self._get_canvas)

    def _get_done(self):
        return self._done

    def _get_canvas(self):
        return self._canvas

    def evaluate(self):
        if self.get_condition() or self.on_init:
            self._canvas = Canvas()
            self.on_init = False
            self._done = True



from bge import logic
from uplogic.nodes import ULActionNode
import bpy


class ULSetEeveeSMAA(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.value = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if self.get_condition():
            print("SMAA has been removed from UPBGE. Delete 'Set SMAA' node to avoid issues.")
            self._done = True
